# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import binascii
import logging
import re
import typing
import uuid
import xml.etree.ElementTree as ET
from copy import copy
from queue import Empty, Queue

from cryptography.hazmat.primitives.padding import PKCS7

from pypsrp._utils import to_bytes, to_string, to_unicode
from pypsrp.complex_objects import (
    ApartmentState,
    Color,
    CommandMetadataCount,
    CommandOrigin,
    ComplexObject,
    Coordinates,
    CultureInfo,
    DictionaryMeta,
    GenericComplexObject,
    HostMethodIdentifier,
    InformationalRecord,
    KeyInfoDotNet,
    ListMeta,
    ObjectMeta,
    ParameterMetadata,
    PipelineResultTypes,
    ProgressRecordType,
    PSCredential,
    PSThreadOptions,
    QueueMeta,
    RemoteStreamOptions,
    SessionStateEntryVisibility,
    Size,
    StackMeta,
)
from pypsrp.exceptions import SerializationError
from pypsrp.messages import (
    DebugRecord,
    ErrorRecordMessage,
    InformationRecord,
    VerboseRecord,
    WarningRecord,
)

log = logging.getLogger(__name__)


class Serializer(object):
    def __init__(self) -> None:
        self.obj_id = 0
        self.obj: typing.Dict = {}
        self.tn_id = 0
        self.tn: typing.Dict = {}

        self.cipher: typing.Any = None
        # Finds C0, C1 and surrogate pairs in a unicode string for us to
        # encode according to the PSRP rules
        self._serial_str = re.compile("[\u0000-\u001f\u007f-\u009f\U00010000-\U0010ffff]")

        # to support surrogate UTF-16 pairs we need to use a UTF-16 regex
        # so we can replace the UTF-16 string representation with the actual
        # UTF-16 byte value and then decode that
        self._deserial_str = re.compile(b"\\x00_\\x00x([\\0\\w]{8})\\x00_")

    def serialize(
        self,
        value: typing.Any,
        metadata: typing.Optional[ObjectMeta] = None,
        parent: typing.Optional[ET.Element] = None,
        clear: bool = True,
    ) -> typing.Optional[ET.Element]:
        """
        Serializes a raw value or class into an XML Element that can be sent
        over to the remote host.

        :param value: The value to serialize
        :param metadata: Any extra metadata to control how to serialize the
            value, if None then the value will be inferred by the type
        :param parent: Whether to append the element onto a parent element
        :param clear: Whether to clear the Obj and TN reference map, this
            should only be True when initially calling serialize
        :return: The XML Element from the serializied value
        """
        if clear:
            self._clear()

        if isinstance(value, ET.Element):
            if metadata is not None and metadata.name is not None:
                value.attrib["N"] = metadata.name

            if parent is not None:
                parent.append(value)

            return value

        metadata = metadata or ObjectMeta()
        if metadata.tag == "*":
            if isinstance(value, TaggedValue):
                metadata.tag = value.tag
                value = value.value
            else:
                metadata.tag = self._get_tag_from_value(value)

        pack_function: typing.Callable[[ObjectMeta, typing.Any], ET.Element] = {  # type: ignore[assignment] # Not sure why
            # primitive types
            "S": lambda m, d: self._serialize_string(d),
            "ToString": lambda d: self._serialize_string(d),
            "C": lambda m, d: str(ord(d)),
            "B": lambda m, d: str(d).lower(),
            "DT": lambda m, d: None,
            "TS": lambda m, d: str(d),
            "By": lambda m, d: str(d),
            "SB": lambda m, d: str(d),
            "U16": lambda m, d: str(d),
            "I16": lambda m, d: str(d),
            "U32": lambda m, d: str(d),
            "I32": lambda m, d: str(d),
            "U64": lambda m, d: str(d),
            "I64": lambda m, d: str(d),
            "Sg": lambda m, d: str(d),
            "Db": lambda m, d: str(d),
            "D": lambda m, d: str(d),
            "BA": lambda m, d: to_string(base64.b64encode(d)),
            "G": lambda m, d: str(d),
            "URI": lambda m, d: self._serialize_string(d),
            "Version": lambda m, d: str(d),
            "XD": lambda m, d: self._serialize_string(d),
            "SBK": lambda m, d: self._serialize_string(d),
            "SS": lambda m, d: self._serialize_secure_string(d),
            "Obj": self._serialize_obj,
            "ObjDynamic": self._serialize_dynamic_obj,
            "LST": self._serialize_lst,
            "IE": self._serialize_ie,
            "QUE": self._serialize_que,
            "STK": self._serialize_stk,
            "DCT": self._serialize_dct,
        }[
            metadata.tag
        ]

        if value is None:
            if metadata.optional:
                return None
            element = ET.Element("Nil")
        else:
            element_value = pack_function(metadata, value)
            if isinstance(element_value, str):
                element = ET.Element(metadata.tag)
                element.text = element_value
            else:
                element = element_value

        if metadata.name is not None:
            element.attrib["N"] = metadata.name

        if parent is not None:
            parent.append(element)

        return element

    def deserialize(
        self,
        element: typing.Union[ET.Element, str],
        metadata: typing.Optional[ObjectMeta] = None,
        clear: bool = True,
    ) -> typing.Any:
        if clear:
            self._clear()

        if isinstance(element, str):
            element_string = element
            try:
                element = ET.fromstring(element)
            except ET.ParseError as err:
                log.warning("Failed to parse data '%s' as XML, return raw xml: %s" % (element_string, str(err)))
                return element_string
        else:
            xml_string = ET.tostring(element, encoding="utf-8", method="xml")
            element_string = to_string(xml_string)

        metadata = metadata or ObjectMeta()
        if metadata.tag == "*":
            metadata.tag = element.tag

        # get the object types so we store the TN Ref ids for later use
        obj_types = self._get_types_from_obj(element)

        # check if it is a primitive object
        unpack_function: typing.Optional[typing.Callable[[ET.Element], typing.Any]] = {
            # Primitive types
            "S": lambda d: self._deserialize_string(d.text),
            "ToString": lambda d: self._deserialize_string(d.text),
            "C": lambda d: chr(int(d.text)),
            "B": lambda d: d.text.lower() == "true",
            "DT": lambda d: d.text,
            "TS": lambda d: d.text,
            "By": lambda d: int(d.text),
            "SB": lambda d: int(d.text),
            "U16": lambda d: int(d.text),
            "I16": lambda d: int(d.text),
            "U32": lambda d: int(d.text),
            "I32": lambda d: int(d.text),
            "U64": lambda d: int(d.text),
            "I64": lambda d: int(d.text),
            "Sg": lambda d: float(d.text),
            "Db": lambda d: float(d.text),
            "D": lambda d: d.text,  # TODO: deserialize this
            "BA": lambda d: base64.b64decode(d.text),
            "G": lambda d: uuid.UUID(d.text),
            "URI": lambda d: self._deserialize_string(d.text),
            "Nil": lambda d: None,
            "Version": lambda d: d.text,
            "XD": lambda d: self._deserialize_string(d.text),
            "SBK": lambda d: self._deserialize_string(d.text),
            "SS": lambda d: self._deserialize_secure_string(d),
            # references an object already deserialized
            "Ref": lambda d: self.obj[d.attrib["RefId"]],
        }.get(element.tag)

        if unpack_function is not None:
            return unpack_function(element)

        # not a primitive object, so try and decode the complex object
        if type(metadata) == ObjectMeta and metadata.object is None:
            structures = {
                "Selected.Microsoft.PowerShell.Commands.GenericMeasureInfo": ObjectMeta(
                    "Obj", object=CommandMetadataCount
                ),
                "System.Array": ListMeta(),
                "System.Collections.ArrayList": ListMeta(),
                "System.Collections.Hashtable": DictionaryMeta(),
                "System.Collections.Generic.List": ListMeta(),
                "System.Collections.Queue": QueueMeta(),
                "System.Collections.Stack": StackMeta(),
                "System.ConsoleColor": ObjectMeta("Obj", object=Color),
                "System.Management.Automation.CommandOrigin": ObjectMeta("Obj", object=CommandOrigin),
                "System.Management.Automation.DebugRecord": ObjectMeta("Obj", object=DebugRecord),
                "System.Management.Automation.ErrorRecord": ObjectMeta("Obj", object=ErrorRecordMessage),
                "System.Management.Automation.Host.Coordinates": ObjectMeta("Obj", object=Coordinates),
                "System.Management.Automation.Host.KeyInfo": ObjectMeta("Obj", object=KeyInfoDotNet),
                "System.Management.Automation.Host.Size": ObjectMeta("Obj", object=Size),
                "System.Management.Automation.InformationalRecord": ObjectMeta("Obj", object=InformationalRecord),
                "System.Management.Automation.InformationRecord": ObjectMeta("Obj", object=InformationRecord),
                "System.Management.Automation.ParameterMetadata": ObjectMeta("Obj", object=ParameterMetadata),
                "System.Management.Automation.ProgressRecordType": ObjectMeta("Obj", object=ProgressRecordType),
                "System.Management.Automation.PSBoundParametersDictionary": DictionaryMeta(),
                "System.Management.Automation.PSCredential": ObjectMeta("Obj", object=PSCredential),
                "System.Management.Automation.PSObject": ObjectMeta("ObjDynamic", object=GenericComplexObject),
                "System.Management.Automation.PSPrimitiveDictionary": DictionaryMeta(),
                "System.Management.Automation.PSTypeName": ObjectMeta("S"),
                "System.Management.Automation.Remoting.RemoteHostMethodId": ObjectMeta(
                    "Obj", object=HostMethodIdentifier
                ),
                "System.Management.Automation.Runspaces.ApartmentState": ObjectMeta("Obj", object=ApartmentState),
                "System.Management.Automation.Runspaces.PipelineResultTypes": ObjectMeta(
                    "Obj", object=PipelineResultTypes
                ),
                "System.Management.Automation.Runspaces.PSThreadOptions": ObjectMeta("Obj", object=PSThreadOptions),
                "System.Management.Automation.Runspaces.RemoteStreamOptions": ObjectMeta(
                    "Obj", object=RemoteStreamOptions
                ),
                "System.Management.Automation.SessionStateEntryVisibility": ObjectMeta(
                    "Obj", object=SessionStateEntryVisibility
                ),
                "System.Management.Automation.VerboseRecord": ObjectMeta("Obj", object=VerboseRecord),
                "System.Management.Automation.WarningRecord": ObjectMeta("Obj", object=WarningRecord),
                "System.Globalization.CultureInfo": ObjectMeta("Obj", object=CultureInfo),
                # Fallback to the GenericComplexObject
                "System.Object": ObjectMeta("ObjDynamic", object=GenericComplexObject),
                # Primitive types
                "System.String": ObjectMeta("S"),
                "System.Char": ObjectMeta("C"),
                "System.Boolean": ObjectMeta("B"),
                "System.DateTime": ObjectMeta("DT"),
                # None: ObjectMeta("TS"), # duration timespan
                "System.Byte": ObjectMeta("By"),
                "System.SByte": ObjectMeta("SB"),
                "System.UInt16": ObjectMeta("U16"),
                "System.Int16": ObjectMeta("I16"),
                "System.UInt32": ObjectMeta("U32"),
                "System.Int32": ObjectMeta("I32"),
                "System.UInt64": ObjectMeta("U64"),
                "System.Int64": ObjectMeta("I64"),
                "System.Single": ObjectMeta("Sg"),
                "System.Double": ObjectMeta("Db"),
                "System.Decimal": ObjectMeta("D"),
                # None: ObjectMeta("BA"), # Byte array base64 encoded
                "System.Guid": ObjectMeta("G"),
                "System.Uri": ObjectMeta("URI"),
                "System.Version": ObjectMeta("Version"),
                "System.Xml.XmlDocument": ObjectMeta("XD"),
                "System.Management.Automation.ScriptBlock": ObjectMeta("SBK"),
                "System.Security.SecureString": ObjectMeta("SS"),
            }

            # fallback to GenericComplexObject if no types were defined
            if metadata.tag == "Obj" and len(obj_types) == 0:
                obj_types = ["System.Object"]

            metadata = None
            for obj_type in obj_types:
                if obj_type.startswith("Deserialized.System."):
                    obj_type = obj_type[13:]

                is_list = False
                if obj_type.endswith("[]"):
                    obj_type = obj_type[0:-2]
                    is_list = True
                elif obj_type.startswith("System.Collections.Generic.List`1[["):
                    list_info = obj_type[35:-1]
                    obj_type = list_info.split(",")[0]
                    is_list = True
                elif obj_type.startswith("System.Collections.ObjectModel.Collection`1[["):
                    list_info = obj_type[45:-1]
                    obj_type = list_info.split(",")[0]
                    is_list = True
                elif obj_type.startswith("System.Collections.ObjectModel.ReadOnlyCollection`1[["):
                    list_info = obj_type[53:-1]
                    obj_type = list_info.split(",")[0]
                    is_list = True
                elif obj_type.startswith("System.Collections.Generic.Dictionary`2[["):
                    dict_meta = obj_type[41:-2].split("],[")
                    key_type = structures.get(dict_meta[0].split(",")[0], ObjectMeta())
                    value_type = structures.get(dict_meta[1].split(",")[0], ObjectMeta())
                    metadata = DictionaryMeta(dict_key_meta=key_type, dict_value_meta=value_type)
                    break

                obj_meta = structures.get(obj_type)
                if obj_meta is not None:
                    metadata = obj_meta
                    if is_list:
                        metadata = ListMeta(list_value_meta=metadata)
                    break

        # we were unable to find the complex object type so just return the
        # element
        obj: typing.Any
        if metadata is None:
            obj = element_string
        elif metadata.tag == "Obj":
            obj = self._deserialize_obj(element, metadata)
        elif metadata.tag == "ObjDynamic":
            obj = self._deserialize_dynamic_obj(element, metadata)
        elif metadata.tag == "LST":
            obj = self._deserialize_lst(element, metadata)
        elif metadata.tag == "QUE":
            obj = self._deserialize_que(element)
        elif metadata.tag == "STK":
            obj = self._deserialize_stk(element)
        elif metadata.tag == "DCT":
            obj = self._deserialize_dct(element)
        else:
            log.warning("Unknown metadata tag type '%s', failed to deserialize object" % metadata.tag)
            obj = element_string

        if element.tag == "Obj":
            self.obj[element.attrib["RefId"]] = obj

        if isinstance(obj, ComplexObject):
            obj._xml = element_string

        return obj

    def _get_tag_from_value(
        self,
        value: typing.Any,
    ) -> str:
        # Get's the XML tag based on the value type, this is a simple list
        # and explicit tagging is recommended.

        value_type = type(value)
        if value_type == int:
            return "I32"
        elif value_type == bool:
            return "B"
        elif value_type == float:
            return "Sg"
        elif value_type == str:
            return "S"
        elif value_type == bytes:
            # This will only occur in Python 3 as a byte string in Python 2 is
            # a str. If users on that platform want a BA then they need to
            # explicitly set the metadata themselves
            return "BA"
        elif value_type == uuid.UUID:
            return "G"
        elif value_type == list:
            return "LST"
        elif value_type == dict:
            return "DCT"
        elif isinstance(value, Queue):
            return "QUE"
        elif isinstance(value, GenericComplexObject):
            return "ObjDynamic"
        elif isinstance(value, ComplexObject):
            return "Obj"
        else:
            # catch all, this probably isn't right but will not throw an
            # error
            return "S"

    def _serialize_obj(
        self,
        metadata: ObjectMeta,
        value: typing.Any,
    ) -> ET.Element:
        obj = ET.Element("Obj", RefId=self._get_obj_id())

        if len(value._types) > 0:
            self._create_tn(obj, value._types)

        to_string_value = value._to_string
        if to_string_value is not None:
            ET.SubElement(obj, "ToString").text = self._serialize_string(to_string_value)

        for attr, property_meta in value._property_sets:
            attr_value = getattr(value, attr)
            self._create_obj(obj, attr_value, meta=property_meta)

        def serialize_prop(parent: str, properties: typing.Tuple[typing.Tuple[str, ObjectMeta], ...]) -> None:
            if len(properties) == 0:
                return
            parent_et = ET.SubElement(obj, parent)
            for attr, property_meta in properties:
                attr_value = getattr(value, attr)
                self._create_obj(parent_et, attr_value, meta=property_meta)

        serialize_prop("MS", value._extended_properties)
        serialize_prop("Props", value._adapted_properties)

        return obj

    def _serialize_dynamic_obj(
        self,
        metadata: ObjectMeta,
        value: typing.Any,
    ) -> ET.Element:
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        self.obj[obj.attrib["RefId"]] = value

        if len(value.types) > 0:
            self._create_tn(obj, value.types)

        if value.to_string is not None:
            ET.SubElement(obj, "ToString").text = self._serialize_string(value.to_string)

        for prop in value.property_sets:
            self._create_obj(obj, prop)

        def set_properties(element: str, prop_name: str) -> None:
            prop_keys = list(getattr(value, prop_name).keys())
            if len(prop_keys) == 0:
                return

            parent = ET.SubElement(obj, element)
            prop_keys.sort()
            for key in prop_keys:
                prop = getattr(value, prop_name)[key]
                self._create_obj(parent, prop, key=key)

        set_properties("MS", "extended_properties")
        set_properties("Props", "adapted_properties")

        return obj

    def _serialize_que(
        self,
        metadata: QueueMeta,
        values: Queue,
    ) -> ET.Element:
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        if not isinstance(metadata, QueueMeta):
            metadata = QueueMeta(name=metadata.name, optional=metadata.optional)
        self._create_tn(obj, metadata.list_types)

        que = ET.SubElement(obj, "QUE")
        while True:
            try:
                value = values.get(block=False)
                self.serialize(value, metadata.list_value_meta, parent=que, clear=False)
            except Empty:
                break

        return obj

    def _serialize_stk(
        self,
        metadata: StackMeta,
        values: typing.List,
    ) -> ET.Element:
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        self._create_tn(obj, metadata.list_types)

        stk = ET.SubElement(obj, "STK")
        while True:
            try:
                value = values.pop()
                self.serialize(value, metadata.list_value_meta, parent=stk, clear=False)
            except IndexError:
                break

        return obj

    def _serialize_ie(
        self,
        metadata: ListMeta,
        values: typing.List,
    ) -> ET.Element:
        return self._serialize_lst(metadata, values, tag="IE")

    def _serialize_lst(
        self,
        metadata: ListMeta,
        values: typing.List,
        tag: str = "LST",
    ) -> ET.Element:
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        if not isinstance(metadata, ListMeta):
            metadata = ListMeta(name=metadata.name, optional=metadata.optional)
        self._create_tn(obj, metadata.list_types)

        lst = ET.SubElement(obj, tag)
        for value in iter(values):
            entry_meta = copy(metadata.list_value_meta)
            self.serialize(value, entry_meta, parent=lst, clear=False)

        return obj

    def _serialize_dct(
        self,
        metadata: DictionaryMeta,
        values: typing.Dict,
    ) -> ET.Element:
        obj = ET.Element("Obj", RefId=self._get_obj_id())
        if not isinstance(metadata, DictionaryMeta):
            metadata = DictionaryMeta(name=metadata.name, optional=metadata.optional)
        self._create_tn(obj, metadata.dict_types)

        dct = ET.SubElement(obj, "DCT")

        # allow dicts to be defined as a tuple so that the order is kept
        iterator: typing.Iterable[typing.Tuple[typing.Any, typing.Any]]
        if isinstance(values, tuple):
            iterator = values
        else:
            iterator = values.items()

        for key, value in iterator:
            en = ET.SubElement(dct, "En")
            key_meta = copy(metadata.dict_key_meta)
            value_meta = copy(metadata.dict_value_meta)
            self.serialize(key, key_meta, parent=en, clear=False)
            self.serialize(value, value_meta, parent=en, clear=False)

        return obj

    def _serialize_string(
        self,
        value: typing.Optional[str],
    ) -> typing.Optional[str]:
        if value is None:
            return None

        def rplcr(matchobj):
            surrogate_char = matchobj.group(0)
            byte_char = to_bytes(surrogate_char, encoding="utf-16-be")
            hex_char = to_unicode(binascii.hexlify(byte_char)).upper()
            hex_split = [hex_char[i : i + 4] for i in range(0, len(hex_char), 4)]

            return "".join(["_x%s_" % i for i in hex_split])

        # before running the translation we need to make sure _ before x is
        # encoded, normally _ isn't encoded except when preceding x
        string_value = to_unicode(value)

        # The MS-PSRP docs don't state this but the _x0000_ matcher is case insensitive so we need to make sure we
        # escape _X as well as _x.
        string_value = re.sub("(?i)_(x)", "_x005F_\\1", string_value)
        string_value = re.sub(self._serial_str, rplcr, string_value)

        return string_value

    def _serialize_secure_string(
        self,
        value: str,
    ) -> str:
        if self.cipher is None:
            raise SerializationError("Cannot generate secure string as cipher is not initialised")

        # convert the string to a UTF-16 byte string as that is what is
        # expected in Windows. If a byte string (native string in Python 2) was
        # passed in, the sender must make sure it is a valid UTF-16
        # representation and not UTF-8 or else the server will fail to decrypt
        # the secure string in most cases
        string_bytes = to_bytes(value, encoding="utf-16-le")

        padder = PKCS7(self.cipher.algorithm.block_size).padder()
        padded_data = padder.update(string_bytes) + padder.finalize()

        encryptor = self.cipher.encryptor()
        ss_value = encryptor.update(padded_data) + encryptor.finalize()
        ss_string = to_string(base64.b64encode(ss_value))

        return ss_string

    def _deserialize_obj(
        self,
        element: ET.Element,
        metadata: ObjectMeta,
    ) -> typing.Any:
        obj = metadata.object()
        self.obj[element.attrib["RefId"]] = obj

        to_string_value = element.find("ToString")
        if to_string_value is not None:
            obj._to_string = self._deserialize_string(to_string_value.text)

        def deserialize_property(prop_tag: str, properties: typing.Tuple[typing.Tuple[str, ObjectMeta], ...]) -> None:
            for attr, property_meta in properties:
                if attr == "invocation_info":
                    a = ""
                property_name = "Unknown"
                property_filter = ""
                if property_meta.name is not None:
                    property_name = property_meta.name
                    property_filter = "[@N='%s']" % property_meta.name

                tags = [property_meta.tag]
                # The below tags are actually seen as Obj in the parent element
                if property_meta.tag in ["DCT", "LST", "IE", "QUE", "STK", "ObjDynamic"]:
                    tags = ["Obj", "Ref"]

                val = None
                for tag in tags:
                    val = element.find("%s%s%s" % (prop_tag, tag, property_filter))
                    if val is not None:
                        break

                if val is None and not property_meta.optional:
                    val = element.find("%sNil%s" % (prop_tag, property_filter))
                    if val is None:
                        obj_name = str(obj) if obj._to_string is not None else "Unknown"
                        err_msg = "Mandatory return value for '%s' was not found on object %s" % (
                            property_name,
                            obj_name,
                        )
                        raise SerializationError(err_msg)
                    val = None
                elif val is not None:
                    val = self.deserialize(val, property_meta, clear=False)

                setattr(obj, attr, val)

        deserialize_property("", obj._property_sets)
        deserialize_property("Props/", obj._adapted_properties)
        deserialize_property("MS/", obj._extended_properties)

        return obj

    def _deserialize_dynamic_obj(
        self,
        element: ET.Element,
        metadata: ObjectMeta,
    ) -> typing.Any:
        obj = metadata.object()
        self.obj[element.attrib["RefId"]] = obj

        for obj_property in element:
            if obj_property.tag == "TN":
                for obj_type in obj_property:
                    obj.types.append(obj_type.text)
                self.tn[obj_property.attrib["RefId"]] = obj.types
            elif obj_property.tag == "TNRef":
                obj.types = self.tn[obj_property.attrib["RefId"]]
            elif obj_property.tag == "Props":
                for adapted_property in obj_property:
                    key = adapted_property.attrib["N"]
                    value = self.deserialize(adapted_property, clear=False)
                    obj.adapted_properties[key] = value
            elif obj_property.tag == "MS":
                for extended_property in obj_property:
                    key = extended_property.attrib["N"]
                    value = self.deserialize(extended_property, clear=False)
                    obj.extended_properties[key] = value
            elif obj_property.tag == "ToString":
                value = self.deserialize(obj_property, clear=False)
                obj.to_string = value
            else:
                value = self.deserialize(obj_property, clear=False)
                obj.property_sets.append(value)

        return obj

    def _deserialize_lst(
        self,
        element: ET.Element,
        metadata: typing.Optional[ObjectMeta] = None,
    ) -> typing.List:
        value_meta = getattr(metadata, "list_value_meta", None)

        list_value = list(self._deserialize_list_values(element, "LST", value_meta=value_meta))

        return list_value

    def _deserialize_que(
        self,
        element: ET.Element,
    ) -> Queue:
        queue: Queue = Queue()

        for entry in self._deserialize_list_values(element, "QUE"):
            queue.put(entry)

        return queue

    def _deserialize_stk(
        self,
        element: ET.Element,
    ) -> typing.List:
        # no native Stack object in Python so just use a list
        stack = list(self._deserialize_list_values(element, "STK"))

        return stack

    def _deserialize_list_values(
        self,
        element: ET.Element,
        list_type: str,
        value_meta: typing.Optional[ObjectMeta] = None,
    ) -> typing.Iterable[typing.Any]:
        entries = element.find(list_type)
        if entries is None:
            return

        for entry in entries:
            entry_value = self.deserialize(entry, metadata=value_meta, clear=False)
            yield entry_value

    def _deserialize_dct(
        self,
        element: ET.Element,
    ) -> typing.Dict:
        dictionary = {}
        entries = element.findall("DCT/En")
        for entry in entries:
            key = entry.find("*[@N='Key']")
            value = entry.find("*[@N='Value']")

            key = self.deserialize(key if key is not None else "", clear=False)
            value = self.deserialize(value if value is not None else "", clear=False)
            dictionary[key] = value

        return dictionary

    def _deserialize_string(
        self,
        value: typing.Optional[str],
    ) -> str:
        if value is None:
            return ""

        def rplcr(matchobj):
            # The matched object is the UTF-16 byte representation of the UTF-8
            # hex string value. We need to decode the byte str to unicode and
            # then unhexlify that hex string to get the actual bytes of the
            # _x****_ value, e.g.
            # group(0) == b"\x00_\x00x\x000\x000\x000\x00A\x00_"
            # group(1) == b"\x000\x000\x000\x00A"
            # unicode (from utf-16-be) == u"000A"
            # returns b"\x00\x0A"
            match_hex = matchobj.group(1)
            hex_string = to_unicode(match_hex, encoding="utf-16-be")
            return binascii.unhexlify(hex_string)

        # need to ensure we start with a unicode representation of the string
        # so that we can get the actual UTF-16 bytes value from that string
        unicode_value = to_unicode(value)
        unicode_bytes = to_bytes(unicode_value, encoding="utf-16-be")
        bytes_value = re.sub(self._deserial_str, rplcr, unicode_bytes)
        return to_unicode(bytes_value, encoding="utf-16-be")

    def _deserialize_secure_string(self, value: ET.Element) -> typing.Union[ET.Element, str]:
        if self.cipher is None:
            # cipher is not set up so we can't decrypt the string, just return
            # the raw element
            return value

        ss_string = base64.b64decode(value.text or "")
        decryptor = self.cipher.decryptor()
        decrypted_bytes = decryptor.update(ss_string) + decryptor.finalize()

        unpadder = PKCS7(self.cipher.algorithm.block_size).unpadder()
        unpadded_bytes = unpadder.update(decrypted_bytes) + unpadder.finalize()
        decrypted_string = to_unicode(unpadded_bytes, "utf-16-le")

        return decrypted_string

    def _clear(self) -> None:
        self.obj_id = 0
        self.obj = {}
        self.tn = {}
        self.tn_id = 0

    def _get_obj_id(self) -> str:
        ref_id = str(self.obj_id)
        self.obj_id += 1
        return ref_id

    def _get_types_from_obj(
        self,
        element: ET.Element,
    ) -> typing.List[str]:
        obj_types = [e.text or "" for e in element.findall("TN/T")]

        if len(obj_types) > 0:
            ref_id = element.find("TN").attrib["RefId"]  # type: ignore[union-attr] # Mandated by the spec
            self.tn[ref_id] = obj_types

        tn_ref = element.find("TNRef")
        if tn_ref is not None:
            ref_id = tn_ref.attrib["RefId"]
            obj_types = self.tn[ref_id]

        return obj_types

    def _create_tn(
        self,
        parent: ET.Element,
        types: typing.List[str],
    ) -> None:
        main_type = types[0]
        ref_id = self.tn.get(main_type, None)
        if ref_id is None:
            ref_id = self.tn_id
            self.tn_id += 1
            self.tn[main_type] = ref_id

            tn = ET.SubElement(parent, "TN", RefId=str(ref_id))
            for type_name in types:
                ET.SubElement(tn, "T").text = type_name
        else:
            ET.SubElement(parent, "TNRef", RefId=str(ref_id))

    def _create_obj(
        self,
        parent: ET.Element,
        obj: typing.Any,
        key: typing.Optional[str] = None,
        meta: typing.Optional[ObjectMeta] = None,
    ) -> None:
        if isinstance(obj, ComplexObject):
            for ref, value in self.obj.items():
                if value == obj:
                    sub_element = ET.SubElement(parent, "Ref", RefId=ref)
                    if key is not None:
                        sub_element.attrib["N"] = key
                    return

        if meta is None:
            meta = ObjectMeta(name=key)
        self.serialize(obj, metadata=meta, parent=parent, clear=False)


class TaggedValue(object):
    def __init__(
        self,
        tag: str,
        value: typing.Any,
    ) -> None:
        self.tag = tag
        self.value = value



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import logging
import struct
import typing
import uuid
import warnings
import xml.etree.ElementTree as ET

from pypsrp._utils import to_string
from pypsrp.complex_objects import (
    ApartmentState,
    CommandType,
    ComplexObject,
    DictionaryMeta,
    ErrorRecord,
    GenericComplexObject,
    HostInfo,
    HostMethodIdentifier,
    InformationalRecord,
    ListMeta,
    ObjectMeta,
    Pipeline,
    ProgressRecordType,
    PSThreadOptions,
    RemoteStreamOptions,
)
from pypsrp.exceptions import SerializationError

if typing.TYPE_CHECKING:
    from pypsrp.serializer import Serializer

log = logging.getLogger(__name__)


class Destination(object):
    # The destination of a PSRP message
    CLIENT = 0x00000001
    SERVER = 0x00000002


class MessageType(object):
    """
    [MS-PSRP] 2.2.1 PowerShell Remoting Protocol Message - MessageType
    https://msdn.microsoft.com/en-us/library/dd303832.aspx

    Identifier of the message contained within a PSRP message
    """

    SESSION_CAPABILITY = 0x00010002
    INIT_RUNSPACEPOOL = 0x00010004
    PUBLIC_KEY = 0x00010005
    ENCRYPTED_SESSION_KEY = 0x00010006
    PUBLIC_KEY_REQUEST = 0x00010007
    CONNECT_RUNSPACEPOOL = 0x00010008
    RUNSPACEPOOL_INIT_DATA = 0x002100B
    RESET_RUNSPACE_STATE = 0x0002100C
    SET_MAX_RUNSPACES = 0x00021002
    SET_MIN_RUNSPACES = 0x00021003
    RUNSPACE_AVAILABILITY = 0x00021004
    RUNSPACEPOOL_STATE = 0x00021005
    CREATE_PIPELINE = 0x00021006
    GET_AVAILABLE_RUNSPACES = 0x00021007
    USER_EVENT = 0x00021008
    APPLICATION_PRIVATE_DATA = 0x00021009
    GET_COMMAND_METADATA = 0x0002100A
    RUNSPACEPOOL_HOST_CALL = 0x00021100
    RUNSPACEPOOL_HOST_RESPONSE = 0x00021101
    PIPELINE_INPUT = 0x00041002
    END_OF_PIPELINE_INPUT = 0x00041003
    PIPELINE_OUTPUT = 0x00041004
    ERROR_RECORD = 0x00041005
    PIPELINE_STATE = 0x00041006
    DEBUG_RECORD = 0x00041007
    VERBOSE_RECORD = 0x00041008
    WARNING_RECORD = 0x00041009
    PROGRESS_RECORD = 0x00041010
    INFORMATION_RECORD = 0x00041011
    PIPELINE_HOST_CALL = 0x00041100
    PIPELINE_HOST_RESPONSE = 0x00041101


class Message(object):
    def __init__(
        self,
        destination: int,
        rpid: typing.Optional[str],
        pid: typing.Optional[str],
        data: typing.Union[ComplexObject, "PipelineOutput", "PublicKeyRequest"],
        serializer: "Serializer",
    ):
        """
        [MS-PSRP] 2.2.1 PowerShell Remoting Protocol Message
        https://msdn.microsoft.com/en-us/library/dd303832.aspx

        Used to contain a PSRP message in the structure required by PSRP.

        :param destination: The Destination of the message
        :param rpid: The uuid representation of the RunspacePool
        :param pid: The uuid representation of the PowerShell pipeline
        :param data: The PSRP Message object
        :param serializer: The serializer object used to serialize the message
            when packing
        """
        self.destination = destination
        self.message_type = data.MESSAGE_TYPE  # type: ignore[union-attr]  # There's no special type right now

        empty_uuid = uuid.UUID(bytes=b"\x00" * 16)
        self.rpid = uuid.UUID(rpid) if rpid is not None else empty_uuid
        self.pid = uuid.UUID(pid) if pid is not None else empty_uuid
        self.data = data
        self._serializer = serializer

    def pack(self) -> bytes:
        msg: typing.Union[ET.Element, bytes]
        if self.message_type == MessageType.PUBLIC_KEY_REQUEST:
            msg = ET.Element("S")
        elif self.message_type == MessageType.END_OF_PIPELINE_INPUT:
            msg = b""
        elif self.message_type == MessageType.PIPELINE_INPUT:
            pipeline_input = typing.cast(PipelineInput, self.data)
            serialized_input = self._serializer.serialize(pipeline_input.data)
            msg = serialized_input if serialized_input is not None else b""
        elif self.message_type == MessageType.CONNECT_RUNSPACEPOOL and (
            self.data.min_runspaces is None and self.data.max_runspaces is None  # type: ignore[union-attr] # Checked with message_type
        ):
            msg = ET.Element("S")
        else:
            serialized_msg = self._serializer.serialize(self.data)
            msg = serialized_msg if serialized_msg is not None else b""

        if not isinstance(msg, bytes):
            message_data = ET.tostring(msg, encoding="utf-8", method="xml")
        else:
            message_data = msg
        log.debug("Packing PSRP message: %s" % to_string(message_data))

        data = struct.pack("<I", self.destination)
        data += struct.pack("<I", self.message_type)

        # .NET stores uuids/guids in bytes in the little endian form
        data += self.rpid.bytes_le
        data += self.pid.bytes_le
        data += message_data

        return data

    @staticmethod
    def unpack(
        data: bytes,
        serializer: "Serializer",
    ) -> "Message":
        destination = struct.unpack("<I", data[0:4])[0]
        message_type = struct.unpack("<I", data[4:8])[0]
        rpid = str(uuid.UUID(bytes_le=data[8:24]))
        pid = str(uuid.UUID(bytes_le=data[24:40]))

        if data[40:43] == b"\xef\xbb\xbf":
            # 40-43 is the UTF-8 BOM which we don't care about
            message_data = to_string(data[43:])
        else:
            message_data = to_string(data[40:])

        log.debug("Unpacking PSRP message of type %d: %s" % (message_type, message_data))

        message_obj = {
            MessageType.SESSION_CAPABILITY: SessionCapability,
            MessageType.INIT_RUNSPACEPOOL: InitRunspacePool,
            MessageType.PUBLIC_KEY: PublicKey,
            MessageType.ENCRYPTED_SESSION_KEY: EncryptedSessionKey,
            MessageType.PUBLIC_KEY_REQUEST: PublicKeyRequest,
            MessageType.SET_MAX_RUNSPACES: SetMaxRunspaces,
            MessageType.SET_MIN_RUNSPACES: SetMinRunspaces,
            MessageType.RUNSPACE_AVAILABILITY: RunspaceAvailability,
            MessageType.RUNSPACEPOOL_STATE: RunspacePoolStateMessage,
            MessageType.CREATE_PIPELINE: CreatePipeline,
            MessageType.GET_AVAILABLE_RUNSPACES: GetAvailableRunspaces,
            MessageType.USER_EVENT: UserEvent,
            MessageType.APPLICATION_PRIVATE_DATA: ApplicationPrivateData,
            MessageType.GET_COMMAND_METADATA: GetCommandMetadata,
            MessageType.RUNSPACEPOOL_HOST_CALL: RunspacePoolHostCall,
            MessageType.RUNSPACEPOOL_HOST_RESPONSE: RunspacePoolHostResponse,
            MessageType.PIPELINE_INPUT: PipelineInput,
            MessageType.END_OF_PIPELINE_INPUT: EndOfPipelineInput,
            MessageType.PIPELINE_OUTPUT: PipelineOutput,
            MessageType.ERROR_RECORD: ErrorRecordMessage,
            MessageType.PIPELINE_STATE: PipelineState,
            MessageType.DEBUG_RECORD: DebugRecord,
            MessageType.VERBOSE_RECORD: VerboseRecord,
            MessageType.WARNING_RECORD: WarningRecord,
            MessageType.PROGRESS_RECORD: ProgressRecord,
            MessageType.INFORMATION_RECORD: InformationRecord,
            MessageType.PIPELINE_HOST_CALL: PipelineHostCall,
            MessageType.PIPELINE_HOST_RESPONSE: PipelineHostResponse,
            MessageType.CONNECT_RUNSPACEPOOL: ConnectRunspacePool,
            MessageType.RUNSPACEPOOL_INIT_DATA: RunspacePoolInitData,
            MessageType.RESET_RUNSPACE_STATE: ResetRunspaceState,
        }[message_type]

        # PIPELINE_OUTPUT is a weird one, it contains the actual output objects
        # not encapsulated so we set it to a dynamic object and the serializer
        # will work out what is best
        message: typing.Union[ComplexObject, "PipelineOutput", "PublicKeyRequest"]
        if message_type == MessageType.PIPELINE_OUTPUT:
            # try to deserialize using our known objects, if that fails then
            # we want to get a generic object at least but raise a warning
            try:
                message_data = serializer.deserialize(message_data)
            except SerializationError as err:
                warnings.warn(
                    "Failed to deserialize msg, trying to deserialize as generic complex object: %s" % str(err)
                )
                meta = ObjectMeta("ObjDynamic", object=GenericComplexObject)
                message_data = serializer.deserialize(message_data, meta)
            message = PipelineOutput()
            message.data = message_data
        elif message_type == MessageType.PIPELINE_INPUT:
            message_data = serializer.deserialize(message_data)
            message = PipelineInput()
            message.data = message_data
        elif message_type == MessageType.PUBLIC_KEY_REQUEST:
            message = PublicKeyRequest()
        else:
            message_meta = ObjectMeta("Obj", object=message_obj)
            message = serializer.deserialize(message_data, message_meta)

        return Message(destination, rpid, pid, message, serializer)


class SessionCapability(ComplexObject):
    MESSAGE_TYPE = MessageType.SESSION_CAPABILITY

    def __init__(
        self,
        protocol_version: typing.Optional[str] = None,
        ps_version: typing.Optional[str] = None,
        serialization_version: typing.Optional[str] = None,
        time_zone: typing.Optional[bytes] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.1 SESSION_CAPABILITY Message
        https://msdn.microsoft.com/en-us/library/dd340636.aspx

        :param protocol_version: The PSRP version
        :param ps_version: The PowerShell version
        :param serialization_version: The serialization version
        :param time_zone: Time Zone information of the host, should be a byte
            string
        """
        super(SessionCapability, self).__init__()
        self._extended_properties = (
            ("protocol_version", ObjectMeta("Version", name="protocolversion")),
            ("ps_version", ObjectMeta("Version", name="PSVersion")),
            ("serialization_version", ObjectMeta("Version", name="SerializationVersion")),
            ("time_zone", ObjectMeta("BA", name="TimeZone", optional=True)),
        )
        self.protocol_version = protocol_version
        self.ps_version = ps_version
        self.serialization_version = serialization_version
        self.time_zone = time_zone


class InitRunspacePool(ComplexObject):
    MESSAGE_TYPE = MessageType.INIT_RUNSPACEPOOL

    def __init__(
        self,
        min_runspaces: typing.Optional[int] = None,
        max_runspaces: typing.Optional[int] = None,
        thread_options: typing.Optional[PSThreadOptions] = None,
        apartment_state: typing.Optional[ApartmentState] = None,
        host_info: typing.Optional[HostInfo] = None,
        application_arguments: typing.Optional[typing.Dict] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.2 INIT_RUNSPACEPOOL Message
        https://msdn.microsoft.com/en-us/library/dd359645.aspx

        :param min_runspaces: The minimum number of runspaces in the pool
        :param max_runspaces: The maximum number of runspaces in the pool
        :param thread_options: Thread options provided by the higher layer
        :param apartment_state: Apartment state provided by the higher layer
        :param host_info: The client's HostInfo details
        :param application_arguments: Application arguments provided by a
            higher layer, stored in the $PSSenderInfo variable in the pool
        """
        super(InitRunspacePool, self).__init__()
        self._extended_properties = (
            ("min_runspaces", ObjectMeta("I32", name="MinRunspaces")),
            ("max_runspaces", ObjectMeta("I32", name="MaxRunspaces")),
            ("thread_options", ObjectMeta("Obj", name="PSThreadOptions", object=PSThreadOptions)),
            ("apartment_state", ObjectMeta("Obj", name="ApartmentState", object=ApartmentState)),
            ("host_info", ObjectMeta("Obj", name="HostInfo", object=HostInfo)),
            (
                "application_arguments",
                DictionaryMeta(
                    name="ApplicationArguments",
                    dict_types=[
                        "System.Management.Automation.PSPrimitiveDictionary",
                        "System.Collections.Hashtable",
                        "System.Object",
                    ],
                ),
            ),
        )
        self.min_runspaces = min_runspaces
        self.max_runspaces = max_runspaces
        self.thread_options = thread_options
        self.apartment_state = apartment_state
        self.host_info = host_info
        self.application_arguments = application_arguments


class PublicKey(ComplexObject):
    MESSAGE_TYPE = MessageType.PUBLIC_KEY

    def __init__(
        self,
        public_key: typing.Optional[str] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.3 PUBLIC_KEY Message
        https://msdn.microsoft.com/en-us/library/dd644859.aspx

        :param public_key: The Base64 encoding of the public key in the PKCS1
            format.
        """
        super(PublicKey, self).__init__()
        self._extended_properties = (("public_key", ObjectMeta("S", name="PublicKey")),)
        self.public_key = public_key


class EncryptedSessionKey(ComplexObject):
    MESSAGE_TYPE = MessageType.ENCRYPTED_SESSION_KEY

    def __init__(self, session_key=None):
        """
        [MS-PSRP] 2.2.2.4 ENCRYPTED_SESSION_KEY Message
        https://msdn.microsoft.com/en-us/library/dd644930.aspx

        :param session_key: The 256-bit key for AES encryption that has been
            encrypted using the public key from the PUBLIC_KEY message using
            the RSAES-PKCS-v1_5 encryption scheme and then Base64 formatted.
        """
        super(EncryptedSessionKey, self).__init__()
        self._extended_properties = (("session_key", ObjectMeta("S", name="EncryptedSessionKey")),)
        self.session_key = session_key


class PublicKeyRequest(object):
    MESSAGE_TYPE = MessageType.PUBLIC_KEY_REQUEST

    def __init__(self) -> None:
        """
        [MS-PSRP] 2.2.2.5 PUBLIC_KEY_REQUEST Message
        https://msdn.microsoft.com/en-us/library/dd644906.aspx
        """
        super(PublicKeyRequest, self).__init__()


class SetMaxRunspaces(ComplexObject):
    MESSAGE_TYPE = MessageType.SET_MAX_RUNSPACES

    def __init__(
        self,
        max_runspaces: typing.Optional[int] = None,
        ci: typing.Optional[int] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.6 SET_MAX_RUNSPACES Message
        https://msdn.microsoft.com/en-us/library/dd304870.aspx

        :param max_runspaces: The maximum number of runspaces
        :param ci: The ci identifier for the CI table
        """
        super(SetMaxRunspaces, self).__init__()
        self._extended_properties = (
            ("max_runspaces", ObjectMeta("I32", name="MaxRunspaces")),
            ("ci", ObjectMeta("I64", name="CI")),
        )
        self.max_runspaces = max_runspaces
        self.ci = ci


class SetMinRunspaces(ComplexObject):
    MESSAGE_TYPE = MessageType.SET_MIN_RUNSPACES

    def __init__(
        self,
        min_runspaces: typing.Optional[int] = None,
        ci: typing.Optional[int] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.7 SET_MIN_RUNSPACES Message
        https://msdn.microsoft.com/en-us/library/dd340570.aspx

        :param max_runspaces: The minimum number of runspaces
        :param ci: The ci identifier for the CI table
        """
        super(SetMinRunspaces, self).__init__()
        self._extended_properties = (
            ("min_runspaces", ObjectMeta("I32", name="MinRunspaces")),
            ("ci", ObjectMeta("I64", name="CI")),
        )
        self.min_runspaces = min_runspaces
        self.ci = ci


class RunspaceAvailability(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACE_AVAILABILITY

    def __init__(self, response=None, ci=None):
        """
        [MS-PSRP] 2.2.2.8 RUNSPACE_AVAILABILITY Message
        https://msdn.microsoft.com/en-us/library/dd359229.aspx

        :param response: The response from the server
        :param ci: The ci identifier for the CI table
        """
        super(RunspaceAvailability, self).__init__()
        self._extended_properties = (
            ("response", ObjectMeta(name="SetMinMaxRunspacesResponse")),
            ("ci", ObjectMeta("I64", name="ci")),
        )
        self.response = response
        self.ci = ci


class RunspacePoolStateMessage(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_STATE

    def __init__(self, state=None, error_record=None):
        """
        [MS-PSRP] 2.2.2.9 RUNSPACEPOOL_STATE Message
        https://msdn.microsoft.com/en-us/library/dd303020.aspx

        :param state: The state of the runspace pool
        :param error_record:
        """
        super(RunspacePoolStateMessage, self).__init__()
        self._extended_properties = (
            ("state", ObjectMeta("I32", name="RunspaceState")),
            ("error_record", ObjectMeta("Obj", optional=True, object=ErrorRecord)),
        )
        self.state = state
        self.error_record = error_record


class CreatePipeline(ComplexObject):
    MESSAGE_TYPE = MessageType.CREATE_PIPELINE

    def __init__(
        self,
        no_input: typing.Optional[bool] = None,
        apartment_state: typing.Optional[ApartmentState] = None,
        remote_stream_options: typing.Optional[RemoteStreamOptions] = None,
        add_to_history: typing.Optional[bool] = None,
        host_info: typing.Optional[HostInfo] = None,
        pipeline: typing.Optional[Pipeline] = None,
        is_nested: typing.Optional[bool] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.10 CREATE_PIPELINE Message
        https://msdn.microsoft.com/en-us/library/dd340567.aspx

        :param no_input: Whether the pipeline will take input
        :param apartment_state: The ApartmentState of the pipeline
        :param remote_stream_options: The RemoteStreamOptions of the pipeline
        :param add_to_history: Whether to add the pipeline being execute to
            the history field of the runspace
        :param host_info: The HostInformation of the pipeline
        :param pipeline: The PowerShell object to create
        :param is_nested: Whether the pipeline is run in nested or
            steppable mode
        """
        super(CreatePipeline, self).__init__()
        self._extended_properties = (
            ("no_input", ObjectMeta("B", name="NoInput")),
            ("apartment_state", ObjectMeta("Obj", name="ApartmentState", object=ApartmentState)),
            ("remote_stream_options", ObjectMeta("Obj", name="RemoteStreamOptions", object=RemoteStreamOptions)),
            ("add_to_history", ObjectMeta("B", name="AddToHistory")),
            ("host_info", ObjectMeta("Obj", name="HostInfo", object=HostInfo)),
            ("pipeline", ObjectMeta("Obj", name="PowerShell", object=Pipeline)),
            ("is_nested", ObjectMeta("B", name="IsNested")),
        )

        self.no_input = no_input
        self.apartment_state = apartment_state
        self.remote_stream_options = remote_stream_options
        self.add_to_history = add_to_history
        self.host_info = host_info
        self.pipeline = pipeline
        self.is_nested = is_nested


class GetAvailableRunspaces(ComplexObject):
    MESSAGE_TYPE = MessageType.GET_AVAILABLE_RUNSPACES

    def __init__(
        self,
        ci: typing.Optional[int] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.11 GET_AVAILABLE_RUNSPACES Message
        https://msdn.microsoft.com/en-us/library/dd357512.aspx

        :param ci: The ci identifier for the CI table
        """
        super(GetAvailableRunspaces, self).__init__()
        self._extended_properties = (("ci", ObjectMeta("I64", name="ci")),)
        self.ci = ci


class UserEvent(ComplexObject):
    MESSAGE_TYPE = MessageType.USER_EVENT

    def __init__(
        self,
        event_id=None,
        source_id=None,
        time=None,
        sender=None,
        args=None,
        data=None,
        computer=None,
        runspace_id=None,
    ):
        """
        [MS-PSRP] 2.2.2.12 USER_EVENT Message
        https://msdn.microsoft.com/en-us/library/dd359395.aspx

        :param event_id:
        :param source_id:
        :param time:
        :param sender:
        :param args:
        :param data:
        :param computer:
        :param runspace_id:
        """
        super(UserEvent, self).__init__()
        self._extended_properties = (
            ("event_id", ObjectMeta("I32", name="PSEventArgs.EventIdentifier")),
            ("source_id", ObjectMeta("S", name="PSEventArgs.SourceIdentifier")),
            ("time", ObjectMeta("DT", name="PSEventArgs.TimeGenerated")),
            ("sender", ObjectMeta(name="PSEventArgs.Sender")),
            ("args", ObjectMeta(name="PSEventArgs.SourceArgs")),
            ("data", ObjectMeta(name="PSEventArgs.MessageData")),
            ("computer", ObjectMeta("S", name="PSEventArgs.ComputerName")),
            ("runspace_id", ObjectMeta("G", name="PSEventArgs.RunspaceId")),
        )
        self.event_id = event_id
        self.source_id = source_id
        self.time = time
        self.sender = sender
        self.args = args
        self.data = data
        self.computer = computer
        self.runspace_id = runspace_id


class ApplicationPrivateData(ComplexObject):
    MESSAGE_TYPE = MessageType.APPLICATION_PRIVATE_DATA

    def __init__(self, data=None):
        """
        [MS-PSRP] 2.2.2.13 APPLICATION_PRIVATE_DATA Message
        https://msdn.microsoft.com/en-us/library/dd644934.aspx

        :param data: A dict that contains data to sent to the PowerShell layer
        """
        super(ApplicationPrivateData, self).__init__()
        self._extended_properties = (("data", DictionaryMeta(name="ApplicationPrivateData")),)
        self.data = data


class GetCommandMetadata(ComplexObject):
    MESSAGE_TYPE = MessageType.GET_COMMAND_METADATA

    def __init__(
        self,
        names: typing.Optional[typing.List[str]] = None,
        command_type: typing.Optional[int] = None,
        namespace: typing.Optional[typing.List[str]] = None,
        argument_list: typing.Optional[typing.List] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.14 GET_COMMAND_METADATA Message
        https://msdn.microsoft.com/en-us/library/ee175985.aspx

        :param names:
        :param command_type:
        :param namespace:
        :param argument_list:
        """
        super(GetCommandMetadata, self).__init__()
        self._extended_properties = (
            (
                "names",
                ListMeta(
                    name="Name",
                    list_value_meta=ObjectMeta("S"),
                    list_types=["System.String[]", "System.Array", "System.Object"],
                ),
            ),
            ("command_type", ObjectMeta(name="CommandType", object=CommandType)),
            ("namespace", ObjectMeta(name="Namespace")),
            ("argument_list", ListMeta(name="ArgumentList")),
        )
        self.names = names
        self.command_type = command_type
        self.namespace = namespace
        self.argument_list = argument_list


class RunspacePoolHostCall(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_HOST_CALL

    def __init__(self, ci=None, mi=None, mp=None):
        """
        [MS-PSRP] 2.2.2.15 RUNSPACE_HOST_CALL Message
        https://msdn.microsoft.com/en-us/library/dd340830.aspx

        :param ci:
        :param mi:
        :param mp:
        """
        super(RunspacePoolHostCall, self).__init__()
        self._extended_properties = (
            ("ci", ObjectMeta("I64", name="ci")),
            ("mi", ObjectMeta("Obj", name="mi", object=HostMethodIdentifier)),
            ("mp", ListMeta(name="mp")),
        )
        self.ci = ci
        self.mi = mi
        self.mp = mp


class RunspacePoolHostResponse(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_HOST_RESPONSE

    def __init__(self, ci=None, mi=None, mr=None, me=None):
        """
        [MS-PSRP] 2.2.2.16 RUNSPACEPOOL_HOST_RESPONSE Message
        https://msdn.microsoft.com/en-us/library/dd358453.aspx

        :param ci:
        :param mi:
        :param mr:
        :param me:
        """
        super(RunspacePoolHostResponse, self).__init__()
        self._extended_properties = (
            ("ci", ObjectMeta("I64", name="ci")),
            ("mi", ObjectMeta("Obj", name="mi", object=HostMethodIdentifier)),
            ("mr", ObjectMeta(name="mr")),
            ("me", ObjectMeta("Obj", name="me", object=ErrorRecord, optional=True)),
        )
        self.ci = ci
        self.mi = mi
        self.mr = mr
        self.me = me


class PipelineInput(ComplexObject):
    MESSAGE_TYPE = MessageType.PIPELINE_INPUT

    def __init__(
        self,
        data: typing.Any = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.17 PIPELINE_INPUT Message
        https://msdn.microsoft.com/en-us/library/dd340525.aspx

        :param data: The data to serialize and send as the input
        """
        super(PipelineInput, self).__init__()
        self.data = data


class EndOfPipelineInput(ComplexObject):
    MESSAGE_TYPE = MessageType.END_OF_PIPELINE_INPUT

    def __init__(self) -> None:
        """
        [MS-PSRP] 2.2.2.18 END_OF_PIPELINE_INPUT Message
        https://msdn.microsoft.com/en-us/library/dd342785.aspx
        """
        super(EndOfPipelineInput, self).__init__()


class PipelineOutput(object):
    MESSAGE_TYPE = MessageType.PIPELINE_OUTPUT

    def __init__(
        self,
        data: typing.Any = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.19 PIPELINE_OUTPUT Message
        https://msdn.microsoft.com/en-us/library/dd357371.aspx
        """
        super(PipelineOutput, self).__init__()
        self.data = data


class ErrorRecordMessage(ErrorRecord):
    MESSAGE_TYPE = MessageType.ERROR_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.20 ERROR_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd342423.aspx

        :param kwargs:
        """
        super(ErrorRecordMessage, self).__init__(**kwargs)


class PipelineState(ComplexObject):
    MESSAGE_TYPE = MessageType.PIPELINE_STATE

    def __init__(self, state=None, error_record=None):
        """
        [MS-PSRP] 2.2.2.21 PIPELINE_STATE Message
        https://msdn.microsoft.com/en-us/library/dd304923.aspx

        :param state: The state of the pipeline
        :param error_record:
        """
        super(PipelineState, self).__init__()
        self._extended_properties = (
            ("state", ObjectMeta("I32", name="PipelineState")),
            ("error_record", ObjectMeta("Obj", name="ExceptionAsErrorRecord", optional=True)),
        )
        self.state = state
        self.error_record = error_record


class DebugRecord(InformationalRecord):
    MESSAGE_TYPE = MessageType.DEBUG_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.22 DEBUG_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd340758.aspx
        """
        super(DebugRecord, self).__init__(**kwargs)
        self._types.insert(0, "System.Management.Automation.DebugRecord")


class VerboseRecord(InformationalRecord):
    MESSAGE_TYPE = MessageType.VERBOSE_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.23 VERBOSE_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd342930.aspx
        """
        super(VerboseRecord, self).__init__(**kwargs)
        self._types.insert(0, "System.Management.Automation.VerboseRecord")


class WarningRecord(InformationalRecord):
    MESSAGE_TYPE = MessageType.WARNING_RECORD

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.24 WARNING_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd303590.aspx
        """
        super(WarningRecord, self).__init__(**kwargs)
        self._types.insert(0, "System.Management.Automation.WarningRecord")


class ProgressRecord(ComplexObject):
    MESSAGE_TYPE = MessageType.PROGRESS_RECORD

    def __init__(
        self,
        activity=None,
        activity_id=None,
        description=None,
        current_operation=None,
        parent_activity_id=None,
        percent_complete=None,
        progress_type=None,
        seconds_remaining=None,
    ):
        """
        [MS-PSRP] 2.2.2.25 PROGRESS_RECORD Message
        https://msdn.microsoft.com/en-us/library/dd340751.aspx

        :param kwargs:
        """
        super(ProgressRecord, self).__init__()
        self._extended_properties = (
            ("activity", ObjectMeta("S", name="Activity")),
            ("activity_id", ObjectMeta("I32", name="ActivityId")),
            ("description", ObjectMeta("S", name="StatusDescription")),
            ("current_operation", ObjectMeta("S", name="CurrentOperation")),
            ("parent_activity_id", ObjectMeta("I32", name="ParentActivityId")),
            ("percent_complete", ObjectMeta("I32", name="PercentComplete")),
            ("progress_type", ObjectMeta("Obj", name="Type", object=ProgressRecordType)),
            ("seconds_remaining", ObjectMeta("I32", name="SecondsRemaining")),
        )
        self.activity = activity
        self.activity_id = activity_id
        self.description = description
        self.current_operation = current_operation
        self.parent_activity_id = parent_activity_id
        self.percent_complete = percent_complete
        self.progress_type = progress_type
        self.seconds_remaining = seconds_remaining


class InformationRecord(ComplexObject):
    MESSAGE_TYPE = MessageType.INFORMATION_RECORD

    def __init__(
        self,
        message_data=None,
        source=None,
        time_generated=None,
        tags=None,
        user=None,
        computer=None,
        pid=None,
        native_thread_id=None,
        managed_thread_id=None,
        write_information_stream=None,
    ):
        """
        [MS-PSRP] 2.2.2.26 INFORMATION_RECORD Message
        https://msdn.microsoft.com/en-us/library/mt224023.aspx

        Only in protocol_version 2.3 and above

        :param kwargs:
        """
        super(InformationRecord, self).__init__()
        self._types = ["System.Management.Automation.InformationRecord", "System.Object"]
        self._extended_properties = (
            ("message_data", ObjectMeta(name="MessageData")),
            ("source", ObjectMeta("S", name="Source")),
            ("time_generated", ObjectMeta("DT", name="TimeGenerated")),
            ("tags", ListMeta(name="Tags", list_value_meta=ObjectMeta("S"))),
            ("user", ObjectMeta("S", name="User")),
            ("computer", ObjectMeta("S", name="Computer")),
            ("pid", ObjectMeta("U32", name="ProcessId")),
            ("native_thread_id", ObjectMeta("U32", name="NativeThreadId")),
            ("managed_thread_id", ObjectMeta("U32", name="ManagedThreadId")),
            ("write_information_stream", ObjectMeta("B", name="WriteInformationStream", optional=True)),
        )
        self.message_data = message_data
        self.source = source
        self.time_generated = time_generated
        self.tags = tags
        self.user = user
        self.computer = computer
        self.pid = pid
        self.native_thread_id = native_thread_id
        self.managed_thread_id = managed_thread_id
        self.write_information_stream = write_information_stream


class PipelineHostCall(RunspacePoolHostCall):
    MESSAGE_TYPE = MessageType.PIPELINE_HOST_CALL
    """
    [MS-PSRP] 2.2.2.27 PIPELINE_HOST_CALL Message
    https://msdn.microsoft.com/en-us/library/dd356915.aspx
    """


class PipelineHostResponse(RunspacePoolHostResponse):
    MESSAGE_TYPE = MessageType.PIPELINE_HOST_RESPONSE
    """
    [MS-PSRP] 2.2.2.28 PIPELINE_HOST_RESPONSE Message
    https://msdn.microsoft.com/en-us/library/dd306168.aspx
    """


class ConnectRunspacePool(ComplexObject):
    MESSAGE_TYPE = MessageType.CONNECT_RUNSPACEPOOL

    def __init__(
        self,
        min_runspaces: typing.Optional[int] = None,
        max_runspaces: typing.Optional[int] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.29 CONNECT_RUNSPACEPOOL Message
        https://msdn.microsoft.com/en-us/library/hh537460.aspx

        :param min_runspaces:
        :param max_runspaces:
        """
        super(ConnectRunspacePool, self).__init__()
        self._extended_properties = (
            ("min_runspaces", ObjectMeta("I32", name="MinRunspaces", optional=True)),
            ("max_runspaces", ObjectMeta("I32", name="MaxRunspaces", optional=True)),
        )
        self.min_runspaces = min_runspaces
        self.max_runspaces = max_runspaces


class RunspacePoolInitData(ComplexObject):
    MESSAGE_TYPE = MessageType.RUNSPACEPOOL_INIT_DATA

    def __init__(self, min_runspaces=None, max_runspaces=None):
        """
        [MS-PSRP] 2.2.2.30 RUNSPACEPOOL_INIT_DATA Message
        https://msdn.microsoft.com/en-us/library/hh537788.aspx

        :param min_runspaces:
        :param max_runspaces:
        """
        super(RunspacePoolInitData, self).__init__()
        self._extended_properties = (
            ("min_runspaces", ObjectMeta("I32", name="MinRunspaces")),
            ("max_runspaces", ObjectMeta("I32", name="MaxRunspaces")),
        )
        self.min_runspaces = min_runspaces
        self.max_runspaces = max_runspaces


class ResetRunspaceState(ComplexObject):
    MESSAGE_TYPE = MessageType.RESET_RUNSPACE_STATE

    def __init__(
        self,
        ci: typing.Optional[int] = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.2.31 RESET_RUNSPACE_STATE Message
        https://msdn.microsoft.com/en-us/library/mt224027.aspx

        :param ci: The call identifier
        """
        super(ResetRunspaceState, self).__init__()
        self._extended_properties = (("ci", ObjectMeta("I64", name="ci")),)
        self.ci = ci



# -*- coding: utf-8 -*-

import logging
import time
import warnings

import pytest
from cryptography.hazmat.primitives.asymmetric import rsa

from pypsrp.complex_objects import (
    Command,
    CommandType,
    ErrorRecord,
    GenericComplexObject,
    ObjectMeta,
    ParameterMetadata,
    PSInvocationState,
    RunspacePoolState,
)
from pypsrp.exceptions import (
    FragmentError,
    InvalidPipelineStateError,
    InvalidPSRPOperation,
    InvalidRunspacePoolStateError,
    WSManFaultError,
)
from pypsrp.messages import (
    Destination,
    ErrorRecordMessage,
    Message,
    MessageType,
    PipelineInput,
    RunspacePoolHostCall,
    RunspacePoolStateMessage,
    UserEvent,
    WarningRecord,
)
from pypsrp.powershell import (
    Fragment,
    Fragmenter,
    PowerShell,
    RunspacePool,
    RunspacePoolWarning,
)
from pypsrp.serializer import Serializer
from pypsrp.wsman import WSMan


def gen_rsa_keypair(public_exponent, key_size):
    # for a passing test against mocked results, we need the same key
    # for each run

    # private key numbers
    d = 21617696977064831737076881102083866512588021127782856037972563073160418492784722157229048881139551797965254106994729989171004895438848876105403145818526588448619012753150859908505079457128264842711189718538434996465875355972895683414261775421352553376878242391757860294122032733780201193753125763293334069627997366690578378643268496972715558088736335242616446993490560907175346811363459760219154246765368454269083276034064771683561116425318584728938420495690895295417261302673227504350398337950587166908640861984752151206832712112769733009288056773792011545187027621461752928443722277452545442134981026071169761569129
    dmp1 = 129490813503009185789974379255770561425157574567257502592607367628590256916193790244618544180042259447193585640234142495635548617466553819050739804131538251753363069954266652888546939369520788464015077722316584767752481491358983245279072645160828476778216654224017198857901235711822454725809127201111535934753
    dmq1 = 48648800397834409354931458481260042874431001732145738308246185682766008281764239801132482123299506611633669689538582946696791307965087683313609603490735622965494394760749284568774366017328170855872209094830510839953297302962573968041079091285470051311421587284299491265676844471173504726139132522742725727613
    iqmp = 39631089624266744377721024140775526581242717587318543319284349361749959043577498327792652579346928867008103676271384098405811933888254575054903735904994321302658687548878487654303384553639708412577672820397125886631056454780398875240593971733447420281054575014889376749144359968455995345232769795875325417317
    p = 174331130742537243408330955079815843963035167715989214203198624738597363753422194561659111132445142920926993058833709875440980890363712769908141629896643845627776407469038077862709547359496382776626050712560846670587394813048350142683947903572416102765283173128172045224943986952459084128294017448217886823253
    q = 141639112913055250372907285405879139487409354087944702421480687298597773645578986399236760591500655922107321654521995138945695777437048045372013886449237695348272152351631535441304559114954968164429007883070066144736528002854651716726263900736377199146896337887655964419309789253989248300176824884184363216819

    # public key numbers
    e = 65537
    n = 24692106711502830011203227021058444318027801046612842012663747949974978593541529463344548800446917862266219189049856550539417324579114258210080798360122994007305091566363663241781504651372764226027210216355916383975880112316706422502404691353489765771310270171473497918954906308690817673196552680498374521519566949221302301125182104198985782657283395055909134373469597836420671163965867038455758131344733786842283328454828820406016508955409107145350345035248825315933976893356673777910251028486191789752627573225093968284278302684745743589192378470115772764709506475265246795419324395050366115533203601201395969892207

    public_numbers = rsa.RSAPublicNumbers(e, n)
    numbers = rsa.RSAPrivateNumbers(p, q, d, dmp1, dmq1, iqmp, public_numbers)
    key = numbers.private_key()

    return key


class RSPoolTest(object):
    def __init__(self):
        self.protocol_version = "2.3"
        self.pipelines = {}


class TestRunspacePool(object):
    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_open_runspace"]], indirect=True)
    def test_psrp_open_runspace(self, wsman_conn):
        runspace_pool = RunspacePool(wsman_conn)
        assert runspace_pool.state == RunspacePoolState.BEFORE_OPEN
        runspace_pool.open()
        assert runspace_pool.application_private_data is not None
        assert runspace_pool.state == RunspacePoolState.OPENED
        runspace_pool.close()
        assert runspace_pool.state == RunspacePoolState.CLOSED

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_set_runspaces"]], indirect=True)
    def test_psrp_set_runspaces(self, wsman_conn):
        runspace_pool = RunspacePool(wsman_conn)
        assert runspace_pool.min_runspaces == 1
        assert runspace_pool.max_runspaces == 1

        # test changing before open works
        runspace_pool.min_runspaces = 2
        runspace_pool.max_runspaces = 2
        assert runspace_pool.min_runspaces == 2
        assert runspace_pool.max_runspaces == 2

        # reset back to one for msg testing
        runspace_pool.min_runspaces = 1
        runspace_pool.max_runspaces = 1

        # open the pool and test changing after open
        try:
            runspace_pool.open()
            actual = runspace_pool.get_available_runspaces()
            assert actual == 1
            runspace_pool.min_runspaces = 1
            runspace_pool.max_runspaces = 5
            actual = runspace_pool.get_available_runspaces()
            assert runspace_pool.min_runspaces == 1
            assert runspace_pool.max_runspaces == 5
            assert actual == 5

            runspace_pool.min_runspaces = 2
            assert runspace_pool.min_runspaces == 2

            runspace_pool.max_runspaces = 5
            assert runspace_pool.max_runspaces == 5

            with pytest.raises(InvalidPSRPOperation) as exc:
                runspace_pool.min_runspaces = -1
            assert str(exc.value) == "Failed to set minimum runspaces"

            with pytest.raises(InvalidPSRPOperation) as exc:
                runspace_pool.max_runspaces = -1
            assert str(exc.value) == "Failed to set maximum runspaces"
        finally:
            runspace_pool.close()

    @pytest.mark.parametrize(
        "wsman_conn",
        # cannot really test in a real life scenario so
        # rely on pre-built responses with a timeout in
        # the reply
        [[False, "test_psrp_key_exchange_timeout"]],
        indirect=True,
    )
    def test_psrp_key_exchange_timeout(self, wsman_conn, monkeypatch):
        monkeypatch.setattr("cryptography.hazmat.primitives.asymmetric.rsa.generate_private_key", gen_rsa_keypair)
        with RunspacePool(wsman_conn, session_key_timeout_ms=500) as pool:
            with pytest.raises(InvalidPSRPOperation) as exc:
                pool.exchange_keys()
            assert str(exc.value) == "Timeout while waiting for key exchange"

    @pytest.mark.parametrize(
        "wsman_conn",
        # due to tests sometimes leaving pools open
        # we are going to mock the data for this one
        [[False, "test_psrp_disconnect_runspaces"]],
        indirect=True,
    )
    def test_psrp_disconnect_runspaces(self, wsman_conn):
        runspace_pool = RunspacePool(wsman_conn)
        runspace_pool.open()
        runspace_pool.disconnect()
        assert runspace_pool.state == RunspacePoolState.DISCONNECTED
        runspace_pool.connect()
        assert runspace_pool.state == RunspacePoolState.OPENED
        runspace_pool.disconnect()

        actual = []
        try:
            runspace_pool2 = RunspacePool(wsman_conn)
            runspace_pool2.open()
            runspace_pool2.disconnect()

            # build a new WSMan object with new unique ID and new requests
            # session
            wsman2 = WSMan(wsman_conn.transport.server)
            wsman2.transport = wsman_conn.transport
            wsman2.transport.session = None
            actual = RunspacePool.get_runspace_pools(wsman2)
            assert len(actual) == 2
            assert actual[0].id == runspace_pool.id or actual[1].id == runspace_pool.id
            assert actual[0].id == runspace_pool2.id or actual[1].id == runspace_pool2.id
            assert actual[0].state == RunspacePoolState.DISCONNECTED
            assert actual[1].state == RunspacePoolState.DISCONNECTED
            for pool in actual:
                pool.connect()
                assert pool.state == RunspacePoolState.OPENED
        finally:
            for pool in actual:
                pool.close()
                assert pool.state == RunspacePoolState.CLOSED

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_application_args"]], indirect=True)
    def test_psrp_application_args(self, wsman_conn):
        app_arguments = {"test_var": "abcdef12345"}

        pool = RunspacePool(wsman_conn)
        pool.open(app_arguments)
        try:
            ps = PowerShell(pool)
            ps.add_script("$PSSenderInfo.ApplicationArguments")
            actual = ps.invoke()
            assert actual[0] == app_arguments
        finally:
            pool.close()

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_no_profile"]], indirect=True)
    def test_psrp_no_profile(self, wsman_conn):
        with RunspacePool(wsman_conn, no_profile=True) as pool:
            ps = PowerShell(pool)
            ps.add_script("$env:USERPROFILE.ToUpperInvariant()")
            actual = ps.invoke()
            assert actual[0] == r"C:\WINDOWS\SYSTEM32\CONFIG\SYSTEMPROFILE"

    @pytest.mark.parametrize(
        "wsman_conn",
        # A real test will take 2 minutes to run, we use a pre-built response to
        # replicate this. Uncommenting the time.sleep line can be used to test
        # this against a real host.
        [[False, "test_psrp_is_alive_invalid_selectors"]],
        indirect=True,
    )
    def test_psrp_is_alive_invalid_selectors(self, wsman_conn):
        with RunspacePool(wsman_conn, idle_timeout=60) as pool:
            assert pool.is_alive() is True
            # This test is pretending 60+ seconds has passed, while 60+ seconds
            # will return a shell in Disconnected mode, we need to wait a bit
            # more for the WSMan service to actually reap the shell and return
            # the invalid selectors message we are tested with here.
            # time.sleep(120)
            assert pool.is_alive() is False

    @pytest.mark.parametrize(
        "wsman_conn",
        # A real test will take 1 minute to run, we use a pre-built response to
        # replicate this. Uncommenting the time.sleep line can be used to test
        # this against a real host.
        [[False, "test_psrp_is_alive_state_disconnected"]],
        indirect=True,
    )
    def test_psrp_is_alive_state_disconnected(self, wsman_conn):
        with RunspacePool(wsman_conn, idle_timeout=60) as pool:
            assert pool.is_alive() is True
            # time.sleep(62)
            assert pool.is_alive() is False

    @pytest.mark.parametrize(
        "wsman_conn",
        # We cannot easily test this in real life as it requires us to kill the
        # WSMan process during the sleep time, so we use a pre-built responses.
        [[False, "test_psrp_is_alive_http_error"]],
        indirect=True,
    )
    def test_psrp_is_alive_http_error(self, wsman_conn):
        with RunspacePool(wsman_conn, idle_timeout=60) as pool:
            assert pool.is_alive(timeout=5) is True
            # For a real test, uncomment and stop WinRM service to replicate
            # a HTTP connection timeout
            # time.sleep(10)
            assert pool.is_alive(timeout=5) is False

    @pytest.mark.parametrize(
        "wsman_conn",
        [[True, "test_psrp_is_alive_other_wsman_error"]],
        indirect=True,
    )
    def test_psrp_is_alive_other_wsman_error(self, wsman_conn):
        with RunspacePool(wsman_conn, idle_timeout=60) as pool:
            pool.shell.wsman.operation_timeout = -1
            with pytest.raises(WSManFaultError) as err:
                pool.is_alive()

            assert (
                str(err.value.reason)
                == "The WS-Management service cannot process the request. The XML contains an invalid timeout field in the SOAP header: PT-1S."
            )

            pool.shell.wsman.operation_timeout = 30

            assert pool.state == RunspacePoolState.OPENED

    @pytest.mark.parametrize(
        "wsman_conn",
        # the commands on the server could differ based on
        # each version, used existing responses instead
        [[False, "test_psrp_get_command_metadata"]],
        indirect=True,
    )
    def test_psrp_get_command_metadata(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            actual_full = pool.get_command_metadata("new-pssession*")
            actual_types = pool.get_command_metadata(["new-*"], command_types=CommandType.FUNCTION)
            actual_namespaced = pool.get_command_metadata("Get-*", namespace=["Microsoft.WSMan.Management"])

        assert len(actual_full) == 3
        assert actual_full[0].name == "New-PSSession"
        assert actual_full[1].name == "New-PSSessionConfigurationFile"
        assert actual_full[2].name == "New-PSSessionOption"

        assert len(actual_types) == 2
        assert actual_types[0].name == "New-Guid"
        assert actual_types[1].name == "New-TemporaryFile"

        assert len(actual_namespaced) == 2
        assert actual_namespaced[0].name == "Get-WSManCredSSP"
        assert actual_namespaced[1].name == "Get-WSManInstance"
        actual = actual_full
        actual.extend(actual_types)
        actual.extend(actual_namespaced)

        for command in actual:
            assert isinstance(command.command_type, CommandType)
            assert isinstance(command.parameters, dict)
            for key, value in command.parameters.items():
                assert isinstance(key, str)
                assert isinstance(value, ParameterMetadata)

    @pytest.mark.parametrize(
        "wsman_conn",
        # reset was only added in v2.3, use existing
        # responses for test
        [[False, "test_psrp_reset_runspace_state"]],
        indirect=True,
    )
    def test_psrp_reset_runspace_state(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script("echo hi")
            ps.invoke(add_to_history=True)

            pool.reset_runspace_state()

            ps = PowerShell(pool)
            ps.add_cmdlet("Get-History")
            actual = ps.invoke()

        # should be empty after clearing history
        assert actual == []

    @pytest.mark.parametrize(
        "wsman_conn",
        # reset was only added in v2.3, use existing
        # responses for test
        [[False, "test_psrp_reset_runspace_state_fail"]],
        indirect=True,
    )
    def test_psrp_reset_runspace_state_fail(self, wsman_conn):
        pool = RunspacePool(wsman_conn)
        pool.open()

        try:
            with pytest.raises(InvalidPSRPOperation) as err:
                pool.reset_runspace_state()
            assert str(err.value) == "Failed to reset runspace state"
        finally:
            pool.close()

    @pytest.mark.parametrize(
        "wsman_conn",
        # JEA endpoint was configured manually to only
        # allow Get-Item and access to the WSMan provider
        [[False, "test_psrp_with_jea_configuration"]],
        indirect=True,
    )
    def test_psrp_with_jea_configuration(self, wsman_conn):
        with RunspacePool(wsman_conn, configuration_name="JEARole") as pool:
            ps = PowerShell(pool)
            ps.add_cmdlet("Get-Item").add_parameter("Path", "WSMan:\\localhost\\Service\\AllowUnencrypted")
            actual = ps.invoke()

            ps_fail = PowerShell(pool)
            ps_fail.add_cmdlet("Get-Item").add_parameter("Path", "C:\\Windows")
            ps_fail.add_statement().add_script(
                "Set-Item -Path WSMan:\\localhost\\Service\\AllowUnencrypted -Value $false"
            )
            actual_fail = ps_fail.invoke()

        assert len(actual) == 1
        assert str(actual[0]) == "Microsoft.WSMan.Management.WSManConfigLeafElement"
        assert actual[0].property_sets[0].adapted_properties["Value"] == "true"
        assert ps.had_errors is False

        assert actual_fail == []
        assert ps_fail.had_errors is True
        assert len(ps_fail.streams.error) == 2
        assert str(ps_fail.streams.error[0]) == "Cannot find a provider with the name 'FileSystem'."
        assert (
            str(ps_fail.streams.error[1]) == "The syntax is not supported by this runspace. This can occur "
            "if the runspace is in no-language mode."
        )

    def test_psrp_connect_already_opened(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        rs.state = RunspacePoolState.OPENED
        rs.connect()

    def test_psrp_connect_invalid_state(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        with pytest.raises(InvalidRunspacePoolStateError) as err:
            rs.connect()
        assert err.value.action == "connect to a disconnected Runspace Pool"
        assert err.value.current_state == RunspacePoolState.BEFORE_OPEN
        assert err.value.expected_state == RunspacePoolState.DISCONNECTED
        assert (
            str(err.value) == "Cannot 'connect to a disconnected Runspace Pool' on the "
            "current state 'BeforeOpen', expecting state(s): 'Disconnected'"
        )

    def test_psrp_disconnect_already_disconnected(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        rs.state = RunspacePoolState.DISCONNECTED
        rs.disconnect()

    def test_psrp_disconnect_invalid_state(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        with pytest.raises(InvalidRunspacePoolStateError) as err:
            rs.disconnect()
        assert err.value.action == "disconnect a Runspace Pool"
        assert err.value.current_state == RunspacePoolState.BEFORE_OPEN
        assert err.value.expected_state == RunspacePoolState.OPENED
        assert (
            str(err.value) == "Cannot 'disconnect a Runspace Pool' on the "
            "current state 'BeforeOpen', expecting state(s): 'Opened'"
        )

    def test_psrp_get_command_meta_invalid_state(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        with pytest.raises(InvalidRunspacePoolStateError) as err:
            rs.get_command_metadata("")
        assert err.value.action == "get command metadata"
        assert err.value.current_state == RunspacePoolState.BEFORE_OPEN
        assert err.value.expected_state == RunspacePoolState.OPENED
        assert (
            str(err.value) == "Cannot 'get command metadata' on the current state "
            "'BeforeOpen', expecting state(s): 'Opened'"
        )

    def test_psrp_open_invalid_state(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        rs.state = RunspacePoolState.DISCONNECTED
        with pytest.raises(InvalidRunspacePoolStateError) as err:
            rs.open()
        assert err.value.action == "open a new Runspace Pool"
        assert err.value.current_state == RunspacePoolState.DISCONNECTED
        assert err.value.expected_state == RunspacePoolState.BEFORE_OPEN
        assert (
            str(err.value) == "Cannot 'open a new Runspace Pool' on the current state "
            "'Disconnected', expecting state(s): 'BeforeOpen'"
        )

    def test_psrp_parse_state_failure(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)
        empty_uuid = "00000000-0000-0000-0000-000000000000"

        state_msg = RunspacePoolStateMessage(RunspacePoolState.OPENED)
        message = Message(0x2, empty_uuid, empty_uuid, state_msg, None)
        rs._process_runspacepool_state(message)
        assert rs.state == RunspacePoolState.OPENED

        excp = ErrorRecord()
        excp._to_string = "error msg"
        state_msg = RunspacePoolStateMessage(RunspacePoolState.BROKEN, excp)
        message = Message(0x2, empty_uuid, empty_uuid, state_msg, None)
        with pytest.raises(InvalidPSRPOperation) as err:
            rs._process_runspacepool_state(message)
        assert str(err.value) == "Received a broken RunspacePoolState message: error msg"

    def test_psrp_reset_state_failure(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)

        # before connecting this should just return None
        rs.reset_runspace_state()

        # when state is not BeforeOpen or Opened then it should fail
        rs.state = RunspacePoolState.BROKEN
        with pytest.raises(InvalidRunspacePoolStateError) as err:
            rs.reset_runspace_state()
        assert err.value.action == "reset RunspacePool state"
        assert err.value.current_state == RunspacePoolState.BROKEN
        assert err.value.expected_state == [RunspacePoolState.BEFORE_OPEN, RunspacePoolState.OPENED]
        assert (
            str(err.value) == "Cannot 'reset RunspacePool state' on the current state "
            "'Broken', expecting state(s): 'BeforeOpen, Opened'"
        )

        rs.state = RunspacePoolState.OPENED
        rs.protocol_version = "2.2"
        with pytest.raises(InvalidPSRPOperation) as err:
            rs.reset_runspace_state()
        assert str(err.value) == "Cannot reset runspace state on protocol versions older than 2.3, actual: 2.2"

    def test_psrp_runspace_host_call_no_host(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)

        msg = Message(Destination.CLIENT, None, None, RunspacePoolHostCall(), None)
        actual = rs._process_runspacepool_host_call(msg)
        assert actual == msg.data

    def test_psrp_process_user_event(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)

        msg = Message(Destination.CLIENT, None, None, UserEvent(), None)
        rs._process_user_event(msg)
        assert len(rs.user_events) == 1
        assert rs.user_events[0] == msg.data

    def test_psrp_warning(self):
        wsman = WSMan("")
        rs = RunspacePool(wsman)

        msg = Message(Destination.CLIENT, None, None, WarningRecord(message="warning msg"), None)
        msg.data._to_string = "warning msg"

        with pytest.warns(RunspacePoolWarning, match="warning msg"):
            rs._process_runspacepool_warning(msg)

    def test_psrp_receive_unknown_message_type(self, caplog):
        wsman = WSMan("")
        rs = RunspacePool(wsman)

        # ERROR_RECORD
        msg_data = b"\x01\x00\x00\x00\x05\x10\x04\x00" + b"\x00" * 32
        msg_data += (
            b'<Obj RefId="0"><TN RefId="0"><T>System.Management.Automation.ErrorRecord</T><T>System.Object</T></TN><ToString>error</ToString>'
            b'<MS><B N="SerializeExtendedInfo">false</B><S N="FullyQualifiedErrorId">ErrorId</S><I32 N="ErrorCategory_Category">1</I32></MS></Obj>'
        )
        fragment = Fragment(0, 0, msg_data, start=True, end=True).pack()

        with caplog.at_level(logging.WARNING, logger="pypsrp.powershell"):
            actual = rs._parse_responses(fragment)

        assert len(actual) == 1
        assert actual[0][0] == 266245
        assert str(actual[0][1].data) == "error"
        assert len(caplog.messages) == 1
        assert caplog.messages[0] == "Unsupported message type '266245' received"


class TestPSRPScenarios(object):
    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_multiple_commands"]], indirect=True)
    def test_psrp_multiple_commands(self, wsman_conn, monkeypatch):
        monkeypatch.setattr("cryptography.hazmat.primitives.asymmetric.rsa.generate_private_key", gen_rsa_keypair)

        with RunspacePool(wsman_conn) as pool:
            assert pool.state == RunspacePoolState.OPENED
            # verify we can still manually call open on an opened pool
            pool.open()

            pool.exchange_keys()
            # exchange keys again and we shouldn't do any ops
            pool.exchange_keys()

            ps = PowerShell(pool)

            # Test out Secure Strings
            sec_string = pool.serialize("Hello World", ObjectMeta("SS"))
            ps.add_cmdlet("Set-Variable")
            ps.add_parameter("Name", "sec_string")
            ps.add_parameter("Value", sec_string)

            ps.add_statement().add_script(
                "[System.Runtime.InteropServices.marshal]"
                "::PtrToStringAuto([System.Runtime.InteropServices.marshal]"
                "::SecureStringToBSTR($sec_string))"
            )
            ps.add_statement().add_cmdlet("ConvertTo-SecureString")
            ps.add_parameter("String", "abc")
            ps.add_parameter("AsPlainText")
            ps.add_parameter("Force")

            # Test out Unicode and complex info
            string_value = "こんにちは - actual_x000A_string\nnewline: %s" % b"\xd8\x01\xdc\x37".decode("utf-16-be")
            ps.add_statement().add_cmdlet("Set-Variable")
            ps.add_parameter("Name", "unicode_string")
            ps.add_parameter("Value", string_value)
            ps.add_statement().add_script("$unicode_string")

            # Arguments
            ps.add_statement().add_cmdlet("cmd.exe").add_argument("/c echo hi")

            # Create Command directly
            command = Command(cmd="whoami.exe", is_script=False)
            ps.add_statement().add_command(command)

            # An integer
            ps.add_statement().add_cmdlet("Set-Variable")
            ps.add_parameter("Name", "integer")
            ps.add_parameter("Value", 123)
            ps.add_statement().add_script("$integer")

            # PSCustomObject
            ps.add_statement().add_cmdlet("Get-Service").add_parameter("Name", "winrm")

            output = ps.invoke()
            assert ps.state == PSInvocationState.COMPLETED

        assert pool.state == RunspacePoolState.CLOSED
        # verify we can still call close on a closed pool
        pool.close()

        assert len(output) == 7
        assert output[0] == "Hello World"
        assert output[1] == "abc"
        assert output[2] == string_value
        assert output[3] == 'hi"'
        # this result differs on whether this is mocked or not
        if type(wsman_conn.transport).__name__ == "TransportFake":
            assert output[4] == "win-nnmu24vvkj0\\vagrant"
        else:
            assert wsman_conn.transport.username.lower() in output[4].lower()
        assert output[5] == 123
        assert isinstance(output[6], GenericComplexObject)
        assert str(output[6]) == "winrm"
        assert output[6].adapted_properties["DisplayName"] == "Windows Remote Management (WS-Management)"
        assert output[6].adapted_properties["ServiceName"] == "winrm"
        assert output[6].extended_properties["Name"] == "winrm"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_multiple_invocations"]], indirect=True)
    def test_psrp_multiple_invocations(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool, PowerShell(pool) as ps:
            ps.add_script("1+1")
            actual1 = ps.invoke()
            assert ps.state == PSInvocationState.COMPLETED
            assert actual1 == [2]

            ps.close()
            ps.clear_streams()

            actual2 = ps.invoke()
            assert ps.state == PSInvocationState.COMPLETED
            assert actual2 == [2]

    @pytest.mark.parametrize(
        "wsman_conn",
        # because we are seeing how the client handles
        # different server protocol versions, we want to
        # use existing responses
        [
            [False, "test_psrp_run_protocol_version_2.1"],
            [False, "test_psrp_run_protocol_version_2.2"],
            [False, "test_psrp_run_protocol_version_2.3"],
        ],
        indirect=True,
    )
    def test_psrp_run_protocol_version(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            if type(wsman_conn.transport).__name__ == "TransportFake":
                expected_version = wsman_conn.transport._test_name.split("_")[-1]
                actual_version = pool.protocol_version
                assert actual_version == expected_version

            ps = PowerShell(pool)
            ps.add_script(
                """begin {
    $DebugPreference = 'Continue'
    Write-Debug "Start Block"
    Write-Error "error"
}
process {
    $input
}
end {
    Write-Debug "End Block"
}"""
            )
            # this tests the merge logic works on v2.1
            ps.merge_error("output")
            actual = ps.invoke(["message 1", 2, ["3", 3]])

            assert len(actual) == 4
            assert str(actual[0]) == "error"
            assert isinstance(actual[0], ErrorRecord)
            assert actual[1] == "message 1"
            assert actual[2] == 2
            assert actual[3] == ["3", 3]
            assert ps.state == PSInvocationState.COMPLETED
            assert ps.had_errors is False
            assert ps.streams.error == []
            assert len(ps.streams.debug) == 2
            assert str(ps.streams.debug[0]) == "Start Block"
            assert str(ps.streams.debug[1]) == "End Block"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_nested_command"]], indirect=True)
    def test_psrp_nested_command(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script("$i = 0; while ($true) { $i++ }")
            ps.begin_invoke()
            time.sleep(0.05)

            nested_ps = ps.create_nested_power_shell()
            nested_ps.add_cmdlet("Get-Variable").add_parameter("Name", "i")
            actual = nested_ps.invoke()
            ps.stop()

        assert actual[0].adapted_properties["Value"] > 0

    @pytest.mark.parametrize(
        "wsman_conn",
        # information stream is not available on all hosts
        # so we just use existing messages
        [[False, "test_psrp_stream_output_invocation"]],
        indirect=True,
    )
    def test_psrp_stream_output_invocation(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)

            script = """$DebugPreference = 'Continue'
            $VerbosePreference = 'Continue'
            Write-Debug 'debug stream'
            Write-Verbose 'verbose stream'
            Write-Error 'error stream'
            Write-Output 'output stream'
            Write-Warning 'warning stream'
            Write-Information 'information stream'
            """

            ps.add_script(script)
            actual = ps.invoke()

        assert ps.state == PSInvocationState.COMPLETED
        assert actual == ["output stream"]
        assert ps.had_errors is False
        assert ps.output == ["output stream"]
        assert len(ps.streams.debug) == 1
        assert str(ps.streams.debug[0]) == "debug stream"
        assert ps.streams.debug[0].invocation
        assert len(ps.streams.error) == 1
        assert str(ps.streams.error[0]) == "error stream"
        assert ps.streams.error[0].invocation
        assert len(ps.streams.information) == 1
        assert ps.streams.information[0].message_data == "information stream"
        assert len(ps.streams.progress) == 1
        assert str(ps.streams.progress[0].progress_type) == "Completed"
        assert len(ps.streams.verbose) == 1
        assert str(ps.streams.verbose[0]) == "verbose stream"
        assert ps.streams.verbose[0].invocation
        assert len(ps.streams.warning) == 1
        assert str(ps.streams.warning[0]) == "warning stream"
        assert ps.streams.warning[0].invocation

    @pytest.mark.parametrize(
        "wsman_conn",
        # information stream is not available on all hosts
        # so we just use existing messages
        [[False, "test_psrp_stream_no_output_invocation"]],
        indirect=True,
    )
    def test_psrp_stream_no_output_invocation(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)

            script = """$DebugPreference = 'Continue'
            $VerbosePreference = 'Continue'
            Write-Debug 'debug stream'
            Write-Verbose 'verbose stream'
            Write-Error 'error stream'
            Write-Output 'output stream'
            Write-Warning 'warning stream'
            Write-Information 'information stream'
            """

            ps.add_script(script)
            actual = ps.invoke(remote_stream_options=0)

        assert ps.state == PSInvocationState.COMPLETED
        assert actual == ["output stream"]
        assert ps.had_errors is False
        assert ps.output == ["output stream"]
        assert len(ps.streams.debug) == 1
        assert str(ps.streams.debug[0]) == "debug stream"
        assert ps.streams.debug[0].invocation is False
        assert len(ps.streams.error) == 1
        assert str(ps.streams.error[0]) == "error stream"
        assert ps.streams.error[0].invocation is False
        assert len(ps.streams.information) == 1
        assert ps.streams.information[0].message_data == "information stream"
        assert len(ps.streams.progress) == 1
        assert str(ps.streams.progress[0].progress_type) == "Completed"
        assert len(ps.streams.verbose) == 1
        assert str(ps.streams.verbose[0]) == "verbose stream"
        assert ps.streams.verbose[0].invocation is False
        assert len(ps.streams.warning) == 1
        assert str(ps.streams.warning[0]) == "warning stream"
        assert ps.streams.warning[0].invocation is False

    @pytest.mark.parametrize(
        "wsman_conn",
        # due to sending the information stream we need
        # to use existing responses as not all servers
        # support this
        [[False, "test_psrp_merge_commands"]],
        indirect=True,
    )
    def test_psrp_merge_commands(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)

            script = """$DebugPreference = 'Continue'
            $VerbosePreference = 'Continue'
            Write-Debug 'debug stream'
            Write-Verbose 'verbose stream'
            Write-Error 'error stream'
            Write-Output 'output stream'
            Write-Warning 'warning stream'
            Write-Information 'information stream'
            """
            ps.add_script(script)
            ps.merge_all("output")
            ps.merge_previous(True)

            ps.add_statement().add_script(script)
            ps.merge_all("output")
            ps.merge_reset()

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                actual = ps.invoke()

            assert len(actual) == 7
            assert len(ps.streams.debug) == 1
            assert len(ps.streams.error) == 1
            assert len(ps.streams.information) == 1
            assert len(ps.streams.progress) == 1
            assert len(ps.streams.verbose) == 1
            assert len(ps.streams.warning) == 1

            assert str(actual[0]) == "debug stream"
            assert str(actual[1]) == "verbose stream"
            assert str(actual[2]) == "error stream"
            assert str(actual[3]) == "output stream"
            assert str(actual[4]) == "warning stream"
            assert str(actual[5]) == "information stream"

            # the 2nd statement should only have the output stream
            assert str(actual[6]) == "output stream"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_error_failed"]], indirect=True)
    def test_psrp_error_failed(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script("$ErrorActionPreference = 'Stop'; Write-Output before; Write-Error error; Write-Output after")
            actual = ps.invoke()

        assert ps.state == PSInvocationState.FAILED
        assert ps.had_errors
        assert actual == ["before"]
        assert len(ps.streams.error) == 1
        assert str(ps.streams.error[0]) == "error"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_with_history"]], indirect=True)
    def test_psrp_with_history(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script("Write-Output 1; Write-Output 2")
            ps.invoke(add_to_history=True)
            ps_hist = PowerShell(pool)
            ps_hist.add_script("Get-History")
            actual = ps_hist.invoke()
        assert len(actual) == 1
        assert actual[0].adapted_properties["CommandLine"] == "Write-Output 1; Write-Output 2"
        assert actual[0].adapted_properties["ExecutionStatus"] == "Completed"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_with_no_history"]], indirect=True)
    def test_psrp_with_no_history(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script("Write-Output 1; Write-Output 2")
            ps.invoke()
            ps_hist = PowerShell(pool)
            ps_hist.add_script("Get-History")
            actual = ps_hist.invoke()
        assert actual == []

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_with_input"]], indirect=True)
    def test_psrp_with_input(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script(
                """begin {
                $DebugPreference = 'Continue'
                Write-Debug "Start Block"
            }
            process {
                $input
            }
            end {
                Write-Debug "End Block"
            }"""
            )
            actual = ps.invoke(["1", 2, {"a": "b"}, ["a", "b"]])

        assert actual == ["1", 2, {"a": "b"}, ["a", "b"]]
        assert str(ps.streams.debug[0]) == "Start Block"
        assert str(ps.streams.debug[1]) == "End Block"

    @pytest.mark.parametrize(
        "wsman_conn",
        # the message size can differ from hosts, we will
        # just use existing responses to get the same
        # scenario each time
        [[False, "test_psrp_small_msg_size"]],
        indirect=True,
    )
    def test_psrp_small_msg_size(self, wsman_conn):
        # we need to set the endpoint for the fake tests to the same length as
        # the one that created the message. This is so the max payload size is
        # the same
        wsman_conn.transport.endpoint = "https://127.0.0.1:55986/wsman"
        wsman_conn.update_max_payload_size()

        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            # there seems to be a bug in the PSRP implementation, I cannot get
            # it to response with a fragment larger than the max allowed so
            # we will just test it receives our large fragments that are split
            ps.add_script(
                """begin {
    $big_var = '%s'
} process {
    $input
} end {
    $big_var[0..19999] -join ""
    $big_var[20000..30000] -join ""
}"""
                % ("a" * 30000)
            )
            actual = ps.invoke("input")
        assert actual[0] == "input"
        assert actual[1] == "a" * 20000
        assert actual[2] == "a" * 10000

    @pytest.mark.parametrize(
        "wsman_conn",
        # so we don't wait 10 seconds in a test we use
        # pre-built responses
        [[False, "test_psrp_long_running_cmdlet"]],
        indirect=True,
    )
    def test_psrp_long_running_cmdlet(self, wsman_conn):
        wsman_conn.operation_timeout = 5

        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_cmdlet("Start-Sleep").add_parameter("Seconds", 10)
            ps.add_statement().add_script("echo hi")
            actual = ps.invoke()
        assert actual[0] == "hi"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_clear_commands"]], indirect=True)
    def test_psrp_clear_command(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.add_script("echo original")
            ps.clear_commands()
            ps.add_script("echo new")
            actual = ps.invoke()
        assert actual[0] == "new"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_receive_failure"]], indirect=True)
    def test_psrp_receive_failure(self, wsman_conn):
        with RunspacePool(wsman_conn) as pool:
            ps = PowerShell(pool)
            ps.state = PSInvocationState.RUNNING
            ps._command_id = ps.id
            pool.pipelines[ps.id] = ps

            with pytest.raises(WSManFaultError) as err:
                ps.end_invoke()
            assert (
                str(err.value.reason) == "The Windows Remote Shell received a request to perform an "
                "operation on a command identifier that does not exist. "
                "Either the command has completed execution or the client "
                "specified an invalid command identifier."
            )

    @pytest.mark.parametrize(
        "wsman_conn",
        # due to tests sometimes leaving pools open
        # we are going to mock the data for this one
        [[False, "test_psrp_disconnected_commands"]],
        indirect=True,
    )
    def test_psrp_disconnected_commands(self, wsman_conn):
        pools = None
        try:
            script = "Write-Output 'a'; Start-Sleep -Seconds 5; Write-Output 'b'"
            pool1 = RunspacePool(wsman_conn)
            pool1.open()
            ps1 = PowerShell(pool1)
            ps1.add_script(script)
            ps1.begin_invoke()
            pool1.disconnect()

            pool2 = RunspacePool(wsman_conn)
            pool2.open()
            ps2 = PowerShell(pool2)
            ps2.add_script(script)
            ps2.begin_invoke()
            pool2.disconnect()

            pools = RunspacePool.get_runspace_pools(wsman_conn)
            assert len(pools) == 2
            assert len(pools[0].pipelines.keys()) == 1
            assert len(pools[1].pipelines.keys()) == 1
            for idx, pool in enumerate(pools):
                pool.connect()
                pipelines = pool.create_disconnected_power_shells()
                for pipeline in pipelines:
                    if idx == 0:
                        pipeline.connect_async()

                        with pytest.raises(InvalidPSRPOperation) as exc:
                            pipeline.create_nested_power_shell()
                        assert (
                            str(exc.value) == "Cannot created a nested PowerShell pipeline "
                            "from an existing pipeline that was connected to"
                            " remotely"
                        )
                        actual = pipeline.end_invoke()
                    else:
                        actual = pipeline.connect()
                    assert actual == ["a", "b"]
        finally:
            if pools is None:
                pools = RunspacePool.get_runspace_pools(wsman_conn)
            for pool in pools:
                pool.connect()
                pool.close()

    def test_add_parameters(self):
        ps = PowerShell(RSPoolTest())
        ps.add_cmdlet("Test-Path")
        ps.add_parameters({"Path": "path", "ItemType": "Leaf"})
        assert len(ps.commands) == 1
        assert ps.commands[0].cmd == "Test-Path"

        # we can't guarantee the dict order so this is the next best thing
        args = ps.commands[0].args
        assert args[0].name == "Path" or args[0].name == "ItemType"
        assert args[1].name == "Path" or args[1].name == "ItemType"

    def test_connect_async_invalid_state(self):
        ps = PowerShell(RSPoolTest())
        with pytest.raises(InvalidPipelineStateError) as err:
            ps.connect()
        assert err.value.action == "connect to a disconnected pipeline"
        assert err.value.current_state == PSInvocationState.NOT_STARTED
        assert err.value.expected_state == PSInvocationState.DISCONNECTED
        assert (
            str(err.value) == "Cannot 'connect to a disconnected pipeline' on the current "
            "state 'NotStarted', expecting state(s): 'Disconnected'"
        )

    def test_close_not_started(self):
        ps = PowerShell(RSPoolTest())
        ps.close()
        assert ps.state == PSInvocationState.NOT_STARTED

    def test_close_invalid_state(self):
        ps = PowerShell(RSPoolTest())
        ps.state = PSInvocationState.RUNNING

        with pytest.raises(InvalidPipelineStateError) as err:
            ps.close()
        assert err.value.action == "close a PowerShell pipeline"
        assert err.value.current_state == PSInvocationState.RUNNING
        assert err.value.expected_state == [
            PSInvocationState.COMPLETED,
            PSInvocationState.STOPPED,
            PSInvocationState.FAILED,
        ]
        assert (
            str(err.value) == "Cannot 'close a PowerShell pipeline' on the current "
            "state 'Running', expecting state(s): 'Completed, Stopped, Failed'"
        )

    def test_psrp_create_nested_invalid_state(self):
        ps = PowerShell(RSPoolTest())
        with pytest.raises(InvalidPipelineStateError) as err:
            ps.create_nested_power_shell()
        assert err.value.action == "create a nested PowerShell pipeline"
        assert err.value.current_state == PSInvocationState.NOT_STARTED
        assert err.value.expected_state == PSInvocationState.RUNNING
        assert (
            str(err.value) == "Cannot 'create a nested PowerShell pipeline' on the current "
            "state 'NotStarted', expecting state(s): 'Running'"
        )

    def test_psrp_create_nested_from_disconnect(self):
        ps = PowerShell(RSPoolTest())
        ps.state = PSInvocationState.RUNNING
        ps._from_disconnect = True
        with pytest.raises(InvalidPSRPOperation) as err:
            ps.create_nested_power_shell()
        assert (
            str(err.value) == "Cannot created a nested PowerShell pipeline from an existing "
            "pipeline that was connected to remotely"
        )

    def test_psrp_begin_invoke_invalid_state(self):
        ps = PowerShell(RSPoolTest())
        ps.state = PSInvocationState.DISCONNECTED
        with pytest.raises(InvalidPipelineStateError) as err:
            ps.begin_invoke()
        assert err.value.action == "start a PowerShell pipeline"
        assert err.value.current_state == PSInvocationState.DISCONNECTED
        assert err.value.expected_state == [
            PSInvocationState.NOT_STARTED,
            PSInvocationState.STOPPED,
            PSInvocationState.COMPLETED,
            PSInvocationState.FAILED,
        ]
        assert (
            str(err.value) == "Cannot 'start a PowerShell pipeline' on the current state "
            "'Disconnected', expecting state(s): 'NotStarted, Stopped, Completed, Failed'"
        )

    def test_psrp_being_invoke_no_commands(self):
        ps = PowerShell(RSPoolTest())
        with pytest.raises(InvalidPSRPOperation) as err:
            ps.begin_invoke()
        assert str(err.value) == "Cannot invoke PowerShell without any commands being set"

    def test_psrp_stop_already_stopped(self):
        ps = PowerShell(RSPoolTest())
        ps.state = PSInvocationState.STOPPED
        ps.stop()
        ps.state = PSInvocationState.STOPPING

    def test_psrp_stop_non_running_pipeline(self):
        ps = PowerShell(RSPoolTest())
        states = [
            PSInvocationState.NOT_STARTED,
            PSInvocationState.COMPLETED,
            PSInvocationState.DISCONNECTED,
            PSInvocationState.FAILED,
        ]
        state_msg = {
            PSInvocationState.NOT_STARTED: "NotStarted",
            PSInvocationState.COMPLETED: "Completed",
            PSInvocationState.DISCONNECTED: "Disconnected",
            PSInvocationState.FAILED: "Failed",
        }

        for state in states:
            with pytest.raises(InvalidPipelineStateError) as err:
                ps.state = state
                ps.stop()

            assert err.value.action == "stop a running pipeline"
            assert err.value.current_state == state
            assert err.value.expected_state == PSInvocationState.RUNNING
            assert (
                str(err.value) == "Cannot 'stop a running pipeline' on the current state '%s',"
                " expecting state(s): 'Running'" % state_msg[state]
            )

    def test_set_merge_to_invalid_to(self):
        ps = PowerShell(RSPoolTest())
        with pytest.raises(InvalidPSRPOperation) as err:
            ps._set_merge_to("error", "fake", None)
        assert str(err.value) == "Invalid merge to option 'fake', valid values none, null, output"

        with pytest.raises(InvalidPSRPOperation) as err:
            ps._set_merge_to("error", "fake", ["option1", "option2"])
        assert str(err.value) == "Invalid merge to option 'fake', valid values option1, option2"

    def test_set_merge_to_invalid_protocol(self):
        ps = PowerShell(RSPoolTest())
        with pytest.raises(InvalidPSRPOperation) as err:
            ps._set_merge_to("error", "output", None, "3.0")
        assert (
            str(err.value) == "Merge option for 'error' is not supported "
            "in the current protocol version 2.3, "
            "minimum version required 3.0"
        )


class TestFragmenter(object):
    def test_fragment_one_one_fragment(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        empty_uuid = "00000000-0000-0000-0000-000000000000"
        pipeline_input = PipelineInput("12")

        actual = fragmenter.fragment(pipeline_input, empty_uuid)
        assert actual == [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>12</S>"
        ]

    def test_fragment_one_multiple_fragments(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        empty_uuid = "00000000-0000-0000-0000-000000000000"
        pipeline_input = PipelineInput("1234")

        actual = fragmenter.fragment(pipeline_input, empty_uuid)
        assert actual == [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>1234</",
            b"\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x01\x02\x00\x00\x00\x02S>",
        ]

    def test_fragment_really_large(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        empty_uuid = "00000000-0000-0000-0000-000000000000"
        pipeline_input = PipelineInput("a" * 60)

        actual = fragmenter.fragment(pipeline_input, empty_uuid)
        assert actual == [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>aaaaaa",
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00"
            b"\x00\x00\x00\x31"
            b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x02"
            b"\x02"
            b"\x00\x00\x00\x09"
            b"aaaaa</S>",
        ]

    def test_fragment_multiple(self):
        serial = Serializer()
        fragmenter = Fragmenter(140, serial)
        empty_uuid = "00000000-0000-0000-0000-000000000000"
        msg1 = PipelineInput("12")
        msg2 = PipelineInput("34")
        msg3 = PipelineInput("567")
        msg4 = PipelineInput("890")

        actual = fragmenter.fragment_multiple([msg1, msg2, msg3, msg4], empty_uuid)
        assert actual == [
            # actual 1 should fit both msg 1 and 2 exactly
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>12</S>"
            b"\x00\x00\x00\x00\x00\x00\x00\x02"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>34</S>",
            # actual 2 should fit msg 3 and the start of msg 4
            b"\x00\x00\x00\x00\x00\x00\x00\x03"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x32"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>567</S>"
            b"\x00\x00\x00\x00\x00\x00\x00\x04"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x30"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>890</",
            # actual 3 should contain the rest of msg 4
            b"\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x01\x02\x00\x00\x00\x02S>",
        ]

    def test_defragment_one_fragment(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        fragments = [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>12</S>"
        ]
        actual = []
        for fragment in fragments:
            actual.extend(fragmenter.defragment(fragment))

        assert len(actual) == 1
        assert actual[0].message_type == MessageType.PIPELINE_INPUT
        assert actual[0].data.data == "12"

    def test_defragment_two_fragments_one_message(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        fragments = [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>1234</",
            b"\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x01\x02\x00\x00\x00\x02S>",
        ]
        actual = []
        for fragment in fragments:
            actual.extend(fragmenter.defragment(fragment))

        assert len(actual) == 1
        assert actual[0].message_type == MessageType.PIPELINE_INPUT
        assert actual[0].data.data == "1234"

    def test_defragment_large_message(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        fragments = [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>aaaaaa",
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00"
            b"\x00\x00\x00\x31"
            b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x02"
            b"\x02"
            b"\x00\x00\x00\x09"
            b"aaaaa</S>",
        ]
        actual = []
        for fragment in fragments:
            actual.extend(fragmenter.defragment(fragment))

        assert len(actual) == 1
        assert actual[0].message_type == MessageType.PIPELINE_INPUT
        assert actual[0].data.data == "a" * 60

    def test_defragment_multiple_fragments_multiple_messages(self):
        serial = Serializer()
        fragmenter = Fragmenter(140, serial)
        fragments = [
            # actual 1 should fit both msg 1 and 2 exactly
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>12</S>"
            b"\x00\x00\x00\x00\x00\x00\x00\x02"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>34</S>",
            # actual 2 should fit msg 3 and the start of msg 4
            b"\x00\x00\x00\x00\x00\x00\x00\x03"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x03"
            b"\x00\x00\x00\x32"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>567</S>"
            b"\x00\x00\x00\x00\x00\x00\x00\x04"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x30"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>890</",
            # actual 3 should contain the rest of msg 4
            b"\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x01\x02\x00\x00\x00\x02S>",
        ]
        actual = []
        for fragment in fragments:
            actual.extend(fragmenter.defragment(fragment))

        assert len(actual) == 4
        assert actual[0].message_type == MessageType.PIPELINE_INPUT
        assert actual[0].data.data == "12"
        assert actual[1].message_type == MessageType.PIPELINE_INPUT
        assert actual[1].data.data == "34"
        assert actual[2].message_type == MessageType.PIPELINE_INPUT
        assert actual[2].data.data == "567"
        assert actual[3].message_type == MessageType.PIPELINE_INPUT
        assert actual[3].data.data == "890"

    def test_defragment_invalid_frag_id(self):
        serial = Serializer()
        fragmenter = Fragmenter(70, serial)
        fragments = [
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x01"
            b"\x00\x00\x00\x31"
            b"\x02\x00\x00\x00"
            b"\x02\x10\x04\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S>aaaaaa",
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x02"
            b"\x02"
            b"\x00\x00\x00\x09"
            b"aaaaa</S>",
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00\x00\x00\x00\x00\x00\x00\x01"
            b"\x00"
            b"\x00\x00\x00\x31"
            b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        ]
        fragmenter.defragment(fragments[0])
        with pytest.raises(FragmentError) as err:
            fragmenter.defragment(fragments[1])
        assert str(err.value) == "Fragment Fragment Id: 2 != Expected Fragment Id: 1"



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import json
import logging
import logging.config
import os
from logging import NullHandler


def _setup_logging(logger: logging.Logger) -> None:
    log_path = os.environ.get("PYPSRP_LOG_CFG", None)

    if log_path is not None and os.path.exists(log_path):  # pragma: no cover
        # log log config from JSON file
        with open(log_path, "rt") as f:
            config = json.load(f)

        logging.config.dictConfig(config)
    else:
        # no logging was provided
        logger.addHandler(NullHandler())


logger = logging.getLogger(__name__)
_setup_logging(logger)

# Contains a list of features, used by external libraries to determine whether
# a new enough pypsrp is installed to support the features it needs
FEATURES = [
    "wsman_locale",
    "wsman_read_timeout",
    "wsman_reconnections",
]
